import torch
from typing import Union, Sequence, Any, Callable, Mapping
from xitorch.debug.modes import is_debug_enabled
from xitorch._core.pure_function import get_pure_function, make_sibling
from xitorch._utils.misc import set_default_option, TensorNonTensorSeparator, \
    TensorPacker, get_method
from xitorch._utils.assertfuncs import assert_fcn_params
from xitorch._impls.integrate.mcsamples.mcmc import mh, mhcustom, dummy1d
from xitorch._docstr.api_docstr import get_methods_docstr

__all__ = ["mcquad"]

def mcquad(
        ffcn: Union[Callable[..., torch.Tensor], Callable[..., Sequence[torch.Tensor]]],
        log_pfcn: Callable[..., torch.Tensor],
        x0: torch.Tensor,
        fparams: Sequence[Any] = [],
        pparams: Sequence[Any] = [],
        bck_options: Mapping[str, Any] = {},
        method: Union[str, Callable, None] = None,
        **fwd_options) -> Union[torch.Tensor, Sequence[torch.Tensor]]:
    r"""
    Performing monte carlo quadrature to calculate the expectation value:

    .. math::
       \mathbb{E}_p[f] = \frac{\int f(\mathbf{x},\theta_f) p(\mathbf{x},\theta_p)
       \ \mathrm{d}\mathbf{x} }{ \int p(\mathbf{x},\theta_p)\ \mathrm{d}\mathbf{x} }

    Arguments
    ---------
    ffcn: Callable
        The function with to be integrated. Its outputs is a tensor or a
        list of tensors. To call the function: ``ffcn(x, *fparams)``
    log_pfcn: Callable
        The natural logarithm of the probability function. The output should be
        a one-element tensor. To call the function: ``log_pfcn(x, *pparams)``
    x0: torch.Tensor
        Tensor with any size as the initial position.
        The call ``ffcn(x0,*fparams)`` must work.
    fparams: list
        Sequence of any other parameters for ``ffcn``.
    pparams: list
        Sequence of any other parameters for ``gfcn``.
    bck_options: dict
        Options for the backward mcquad operation. Unspecified fields will be
        taken from ``fwd_options``.
    method: str or callable or None
        Monte Carlo quadrature method. If None, it will choose ``"mh"``.
    **fwd_options: dict
        Method-specific options (see method section below).

    Returns
    -------
    torch.Tensor or a list of torch.Tensor
        The expectation values of the function ``ffcn`` over the space of ``x``.
        If the output of ``ffcn`` is a list, then this is also a list.
    """
    if method is None:
        method = "mh"
    return _mcquad(ffcn, log_pfcn, x0, None, None, fparams, pparams,
                   method, bck_options, **fwd_options)

def _mcquad(ffcn, log_pfcn, x0, xsamples, wsamples, fparams, pparams, method,
            bck_options, **fwd_options):
    # this is mcquad with an additional xsamples argument, to prevent xsamples being set by users

    if is_debug_enabled():
        assert_fcn_params(ffcn, (x0, *fparams))
        assert_fcn_params(log_pfcn, (x0, *pparams))

    # check if ffcn produces a list / tuple
    out = ffcn(x0, *fparams)
    is_tuple_out = isinstance(out, list) or isinstance(out, tuple)

    # get the pure functions
    pure_ffcn = get_pure_function(ffcn)
    pure_logpfcn = get_pure_function(log_pfcn)
    nfparams = len(fparams)
    npparams = len(pparams)
    fobjparams = pure_ffcn.objparams()
    pobjparams = pure_logpfcn.objparams()
    nf_objparams = len(fobjparams)

    if is_tuple_out:
        packer = TensorPacker(out)

        @make_sibling(pure_ffcn)
        def pure_ffcn2(x, *fparams):
            y = pure_ffcn(x, *fparams)
            return packer.flatten(y)
        res = _MCQuad.apply(pure_ffcn2, pure_logpfcn, x0, None, None,
                            method, fwd_options, bck_options,
                            nfparams, nf_objparams, npparams, *fparams, *fobjparams, *pparams, *pobjparams)
        return packer.pack(res)
    else:
        return _MCQuad.apply(pure_ffcn, pure_logpfcn, x0, None, None,
                             method, fwd_options, bck_options,
                             nfparams, nf_objparams, npparams, *fparams, *fobjparams, *pparams, *pobjparams)

class _MCQuad(torch.autograd.Function):
    @staticmethod
    def forward(ctx, ffcn, log_pfcn, x0, xsamples, wsamples,
                method, fwd_options, bck_options,
                nfparams, nf_objparams, npparams, *all_fpparams):
        # set up the default options
        config = fwd_options
        ctx.bck_config = set_default_option(config, bck_options)

        # split the parameters
        fparams    = all_fpparams[:nfparams]
        fobjparams = all_fpparams[nfparams:nfparams + nf_objparams]
        pparams    = all_fpparams[nfparams + nf_objparams:nfparams + nf_objparams + npparams]
        pobjparams = all_fpparams[nfparams + nf_objparams + npparams:]

        # select the method for the sampling
        if xsamples is None:
            methods = {
                "mh": mh,
                "_dummy1d": dummy1d,
                "mhcustom": mhcustom,
            }
            method_fcn = get_method("mcquad", methods, method)
            xsamples, wsamples = method_fcn(log_pfcn, x0, pparams, **config)
        epf = _integrate(ffcn, xsamples, wsamples, fparams)

        # save parameters for backward calculations
        ctx.xsamples = xsamples
        ctx.wsamples = wsamples
        ctx.ffcn = ffcn
        ctx.log_pfcn = log_pfcn
        ctx.fparam_sep = TensorNonTensorSeparator((*fparams, *fobjparams))
        ctx.pparam_sep = TensorNonTensorSeparator((*pparams, *pobjparams))
        ctx.nfparams = len(fparams)
        ctx.npparams = len(pparams)
        ctx.method = method

        # save for backward
        ftensor_params = ctx.fparam_sep.get_tensor_params()
        ptensor_params = ctx.pparam_sep.get_tensor_params()
        ctx.nftensorparams = len(ftensor_params)
        ctx.nptensorparams = len(ptensor_params)
        ctx.save_for_backward(epf, *ftensor_params, *ptensor_params)

        return epf

    @staticmethod
    def backward(ctx, grad_epf):
        # restore the parameters
        alltensors = ctx.saved_tensors
        nftensorparams = ctx.nftensorparams
        nptensorparams = ctx.nptensorparams
        epf = alltensors[0]
        ftensor_params = alltensors[1:1 + nftensorparams]
        ptensor_params = alltensors[1 + nftensorparams:]
        fptensor_params = alltensors[1:]

        # get the parameters and the object parameters
        nfparams = ctx.nfparams
        npparams = ctx.npparams
        fall_params = ctx.fparam_sep.reconstruct_params(ftensor_params)
        pall_params = ctx.pparam_sep.reconstruct_params(ptensor_params)
        fparams = fall_params[:nfparams]
        fobjparams = fall_params[nfparams:]
        pparams = pall_params[:npparams]
        pobjparams = pall_params[npparams:]

        # get other things from the forward
        ffcn = ctx.ffcn
        log_pfcn = ctx.log_pfcn
        xsamples = ctx.xsamples
        wsamples = ctx.wsamples
        grad_enabled = torch.is_grad_enabled()

        def function_wrap(fcn, param_sep, nparams, x, tensor_params):
            all_params = param_sep.reconstruct_params(tensor_params)
            params = all_params[:nparams]
            objparams = all_params[nparams:]
            with fcn.useobjparams(objparams):
                f = fcn(x, *params)
            return f

        def aug_function(x, *grad_and_fptensor_params):
            local_grad_enabled = torch.is_grad_enabled()
            grad_epf = grad_and_fptensor_params[0]
            epf = grad_and_fptensor_params[1]
            fptensor_params = grad_and_fptensor_params[2:]
            ftensor_params = fptensor_params[:nftensorparams]
            ptensor_params = fptensor_params[nftensorparams:]
            with torch.enable_grad():
                # if graph is constructed, then fptensor_params is a clone of
                # fptensor_params from outside, therefore, it needs to be put
                # in the pure function's objects (that's what function_wrap does)
                if grad_enabled:
                    fout = function_wrap(ffcn, ctx.fparam_sep, nfparams, x, ftensor_params)
                    pout = function_wrap(log_pfcn, ctx.pparam_sep, npparams, x, ptensor_params)
                # if graph is not constructed, then fptensor_params in this
                # function *is* fptensor_params in the outside, so we can
                # just use fparams and pparams from the outside
                else:
                    fout = ffcn(x, *fparams)
                    pout = log_pfcn(x, *pparams)
            # derivative of fparams
            dLdthetaf = []
            if len(ftensor_params) > 0:
                dLdthetaf = torch.autograd.grad(fout, ftensor_params,
                                                grad_outputs=grad_epf,
                                                retain_graph=True,
                                                create_graph=local_grad_enabled)
            # derivative of pparams
            dLdthetap = []
            if len(ptensor_params) > 0:
                dLdef = torch.dot((fout - epf).reshape(-1), grad_epf.reshape(-1))
                dLdthetap = torch.autograd.grad(pout, ptensor_params,
                                                grad_outputs=dLdef.reshape(pout.shape),
                                                retain_graph=True,
                                                create_graph=local_grad_enabled)
            # combine the states needed for backward
            outs = (
                *dLdthetaf,
                *dLdthetap,
            )
            return outs

        if grad_enabled:
            fptensor_params_copy = [y.clone().requires_grad_() for y in fptensor_params]
        else:
            fptensor_params_copy = fptensor_params

        aug_epfs = _mcquad(aug_function, log_pfcn,
                           x0=xsamples[0],  # unused because xsamples is set
                           xsamples=xsamples,
                           wsamples=wsamples,
                           fparams=(grad_epf, epf, *fptensor_params_copy),
                           pparams=pparams,
                           method=ctx.method,
                           bck_options=ctx.bck_config,
                           **ctx.bck_config)
        dLdthetaf = aug_epfs[:nftensorparams]
        dLdthetap = aug_epfs[nftensorparams:]

        # combine the gradient for all fparams
        dLdfnontensor = [None for _ in range(ctx.fparam_sep.nnontensors())]
        dLdpnontensor = [None for _ in range(ctx.pparam_sep.nnontensors())]
        dLdtf = ctx.fparam_sep.reconstruct_params(dLdthetaf, dLdfnontensor)
        dLdtp = ctx.pparam_sep.reconstruct_params(dLdthetap, dLdpnontensor)
        return (None, None, None, None, None, None, None, None, None, None, None,
                *dLdtf, *dLdtp)

def _integrate(ffcn, xsamples, wsamples, fparams):
    nsamples = len(xsamples)
    res = 0.0
    for x, w in zip(xsamples, wsamples):
        res = res + ffcn(x, *fparams) * w
    return res


# docstring completion
mcquad.__doc__ = get_methods_docstr(mcquad, [mh, mhcustom])



import torch
from dataclasses import dataclass
from xitorch._core.packer import Packer

@dataclass
class MyObj:
    a: torch.Tensor
    b: torch.Tensor
    c: int

def test_packer_tensor():
    # test if the object to be packed is a tensor
    a = torch.tensor(1.5)
    b = torch.tensor(2.0)
    b2 = torch.tensor([1.0, 1.0])
    b3 = torch.tensor([4.0])
    packer = Packer(a)

    for unique in [True, False]:
        print(unique)

        # test get tensors
        tensors = packer.get_param_tensor_list(unique=unique)
        assert len(tensors) == 1
        assert tensors[0] is a

        # test set tensors and check if the packer's state is unchanged
        obj = packer.construct_from_tensor_list([b], unique=unique)
        assert obj is b
        assert packer.get_param_tensor_list(unique=unique)[0] is a

        # get the single param tensor
        ptensor = packer.get_param_tensor(unique=unique)
        assert ptensor is a

        # set tensor from a single param
        obj = packer.construct_from_tensor(b, unique=unique)
        assert obj is b
        assert packer.get_param_tensor(unique=unique) is a

        # test if the input has mismatched numel
        try:
            obj = packer.construct_from_tensor(b2, unique=unique)
            assert False
        except RuntimeError as e:
            pass

        # test if the input has mismatched shapes
        try:
            obj = packer.construct_from_tensor(b3, unique=unique)
            assert False
        except RuntimeError as e:
            pass

def test_packer_list_tensors():
    # test if the object is a list of tensors
    a = torch.tensor(1.0)
    b = torch.tensor(2.0)
    obj0 = [a, a, b, a]
    new_tlist = [b, b, a, b]
    packer = Packer(obj0)

    # get tensors
    tensors = packer.get_param_tensor_list(unique=False)
    assert len(tensors) == len(obj0)
    for i in range(len(tensors)):
        assert tensors[i] is obj0[i]

    # get unique tensors (also test if the default is unique=True)
    tensors = packer.get_param_tensor_list()
    assert len(tensors) == 2
    assert tensors[0] == a
    assert tensors[1] == b

    # construct the object from new list
    obj1 = packer.construct_from_tensor_list(new_tlist, unique=False)
    assert len(obj1) == len(obj0)
    assert obj1[0] is b
    assert obj1[1] is b
    assert obj1[2] is a
    assert obj1[3] is b

    # construct from the unique list
    obj2 = packer.construct_from_tensor_list([b, a])
    assert len(obj2) == len(obj0)
    assert obj2[0] is b
    assert obj2[1] is b
    assert obj2[2] is a
    assert obj2[3] is b

    # test immutability by repeating one of the get params test above
    tensors = packer.get_param_tensor_list()
    assert len(tensors) == 2
    assert tensors[0] == a
    assert tensors[1] == b

    # get one tensor as a param
    tparam = packer.get_param_tensor(unique=False)
    assert tparam.numel() == 4
    assert torch.allclose(tparam, torch.tensor([1., 1., 2., 1.]))
    tparam = packer.get_param_tensor()
    assert tparam.numel() == 2
    assert torch.allclose(tparam, torch.tensor([1., 2.]))

    # set the one tensor param
    obj1 = packer.construct_from_tensor(torch.tensor([2., 3., 1., 4.]), unique=False)
    assert len(obj1) == 4
    assert float(obj1[0].item()) == 2.0
    assert float(obj1[1].item()) == 3.0
    assert float(obj1[2].item()) == 1.0
    assert float(obj1[3].item()) == 4.0

    obj1 = packer.construct_from_tensor(torch.tensor([2., 3.]), unique=True)
    assert len(obj1) == 4
    assert float(obj1[0].item()) == 2.0
    assert float(obj1[1].item()) == 2.0
    assert float(obj1[2].item()) == 3.0
    assert float(obj1[3].item()) == 2.0

    # test immutability
    obj0.append(a)  # change the object, and see if packer changes or not
    tparam = packer.get_param_tensor(unique=False)
    assert tparam.numel() == 4
    assert torch.allclose(tparam, torch.tensor([1., 1., 2., 1.]))
    tparam = packer.get_param_tensor()
    assert tparam.numel() == 2
    assert torch.allclose(tparam, torch.tensor([1., 2.]))

    # test shape mismatch
    try:
        obj2 = packer.construct_from_tensor_list([b, a, b, b.unsqueeze(-1)], unique=False)
        assert False
    except RuntimeError:
        pass

def test_packer_null():
    # test if the packer object contains no tensors
    a = [1, 2, 3]
    packer = Packer(a)

    assert packer.get_param_tensor(unique=False) is None
    assert packer.get_param_tensor(unique=True) is None
    assert packer.get_param_tensor_list(unique=False) == []
    assert packer.get_param_tensor_list(unique=True) == []

    obj = packer.construct_from_tensor(None)
    assert obj == a
    obj2 = packer.construct_from_tensor_list([])
    assert obj2 == a

def test_packer_complex_structure():
    a = torch.tensor([1., 2.]).requires_grad_()
    b = torch.tensor([4.])
    c = torch.cat((a, b))
    a2 = a * 2
    b2 = b * 2
    c2 = c * 2
    a3 = a * 3
    myobj = MyObj(a=a, b=b, c=3)
    obj0 = {
        "lst": [myobj, c],
        "a": a,
    }
    packer = Packer(obj0)

    # test if error is raised to construct before calling get tensors
    try:
        tparam_new = torch.tensor([1., 1., 1., 1., 1., 1.])  # numel == 6 for unique
        packer.construct_from_tensor(tparam_new)
        assert False
    except RuntimeError as e:
        assert "Please execute" in str(e)

    try:
        tparams = [a2, b2, c2, a2]
        packer.construct_from_tensor_list(tparams, unique=False)
        assert False
    except RuntimeError as e:
        assert "Please execute" in str(e)

    # get tensors
    tensors = packer.get_param_tensor_list(unique=False)
    assert len(tensors) == 4  # (a, b, c, a)
    assert tensors[0] is a
    assert tensors[1] is b
    assert tensors[2] is c
    assert tensors[3] is a
    tensors = packer.get_param_tensor_list()
    assert len(tensors) == 3  # (a, b, c)

    tparam = packer.get_param_tensor(unique=False)
    assert tparam.numel() == (a.numel() * 2 + b.numel() + c.numel())
    tparam = packer.get_param_tensor()
    assert tparam.numel() == (a.numel() + b.numel() + c.numel())

    # test construct object
    obj1 = packer.construct_from_tensor_list([a2, b2, c2, a3], unique=False)
    assert obj1["a"] is a3
    assert obj1["lst"][1] is c2
    assert obj1["lst"][0].a is a2
    assert obj1["lst"][0].b is b2
    assert obj1["lst"][0].c == 3
    myobj.c = 4  # test immutability
    assert obj0["lst"][0].c == 4  # make sure the obj0 is changed
    # now make sure the object inside packer is not changed
    obj1 = packer.construct_from_tensor_list([a2, b2, c2])
    assert obj1["a"] is a2
    assert obj1["lst"][1] is c2
    assert obj1["lst"][0].a is a2
    assert obj1["lst"][0].b is b2
    assert obj1["lst"][0].c == 3

    # test immutability of the packer object
    tensors = packer.get_param_tensor_list(unique=False)
    assert len(tensors) == 4  # (a, b, c, a)
    assert tensors[0] is a
    assert tensors[1] is b
    assert tensors[2] is c
    assert tensors[3] is a

    # construct object from a single parameter
    tparam = packer.get_param_tensor()
    obj1 = packer.construct_from_tensor(tparam * 4)
    assert torch.allclose(obj1["a"], a * 4)
    assert torch.allclose(obj1["lst"][1], c * 4)
    assert torch.allclose(obj1["lst"][0].a, a * 4)
    assert torch.allclose(obj1["lst"][0].b, b * 4)
    assert obj1["lst"][0].c == 3
    # make sure the parameters in obj1 is in the correct graph with a
    grad_a, = torch.autograd.grad(obj1["a"].sum(), a, retain_graph=True)
    assert torch.allclose(grad_a, grad_a * 0 + 4)

    # test error mismatch numel
    try:
        tparam_new = torch.tensor([1., 1., 1., 1., 1., 1., 1., 1., 1., 1.])
        packer.construct_from_tensor(tparam_new)
        assert False
    except RuntimeError as e:
        assert "match" in str(e)



import warnings
import torch
from xitorch import LinearOperator, EditableModule
from xitorch.linalg import solve, symeig

class LinOpWithoutGetParamNames(LinearOperator):
    def __init__(self, mat, is_hermitian=False):
        super(LinOpWithoutGetParamNames, self).__init__(
            shape=mat.shape,
            is_hermitian=is_hermitian,
            dtype=mat.dtype,
            device=mat.device
        )
        self.mat = mat
        self.implemented_methods = []

    def _mv(self, x):
        return torch.matmul(self.mat, x.unsqueeze(-1)).squeeze(-1)

class LinOpWithoutInit(LinearOperator):
    def __init__(self, mat, is_hermitian=False):
        self.mat = mat
        self.implemented_methods = []

    def _mv(self, x):
        return torch.matmul(self.mat, x.unsqueeze(-1)).squeeze(-1)

class BaseLinOp(LinearOperator):
    def __init__(self, mat, is_hermitian=False):
        super(BaseLinOp, self).__init__(
            shape=mat.shape,
            is_hermitian=is_hermitian,
            dtype=mat.dtype,
            device=mat.device
        )
        self.mat = mat
        self.implemented_methods = []

    def _getparamnames(self, prefix=""):
        return [prefix + "mat"]

class LinOpWithoutMv(BaseLinOp):
    # LinearOperator where only nothing is implemented (should produce an error)
    def __init__(self, mat, is_hermitian=False):
        super(LinOpWithoutMv, self).__init__(mat, is_hermitian)
        self.implemented_methods = []

class LinOp1(BaseLinOp):
    # LinearOperator where only ._mv is implemented (bare minimum)
    def __init__(self, mat, is_hermitian=False):
        super(LinOp1, self).__init__(mat, is_hermitian)
        self.implemented_methods = ["_mv"]

    def _mv(self, x):
        return torch.matmul(self.mat, x.unsqueeze(-1)).squeeze(-1)

class LinOp2(BaseLinOp):
    # LinearOperator where ._mv and ._rmv are implemented
    def __init__(self, mat, is_hermitian=False):
        super(LinOp2, self).__init__(mat, is_hermitian)
        self.implemented_methods = ["_mv", "_rmv"]

    def _mv(self, x):
        return torch.matmul(self.mat, x.unsqueeze(-1)).squeeze(-1)

    def _rmv(self, x):
        return torch.matmul(self.mat.transpose(-2, -1), x.unsqueeze(-1)).squeeze(-1)

class NotLinOp1(BaseLinOp):
    # LinearOperator where only ._mv is implemented (bare minimum)
    def __init__(self, mat, is_hermitian=False):
        super(NotLinOp1, self).__init__(mat, is_hermitian)
        self.implemented_methods = ["_mv"]

    def _mv(self, x):
        return torch.matmul(self.mat, x.unsqueeze(-1)).squeeze(-1) + 1

def test_linop0_err():
    mat = torch.rand((3, 1, 2))
    try:
        linop = LinOpWithoutMv(mat)
        msg = ("A RuntimeError must be raised when creating a "
               "LinearOperator without ._mv()")
        assert False, msg
    except RuntimeError:
        pass

def test_linop_no_getparamnames_err():
    mat = torch.rand((3, 2, 2))
    linop = LinOpWithoutGetParamNames(mat)
    x = torch.rand(2, 4)
    b = linop.mm(x)
    with torch.no_grad():
        solve(linop, b)
    try:
        with torch.enable_grad():
            solve(linop, b)
        msg = ("A RuntimeError must be raised when using a LinearOperator "
               "without ._getparamnames() in xitorch functions")
        assert False, msg
    except RuntimeError:
        pass

def test_linop_no_init_err():
    mat = torch.rand((3, 1, 2))
    x = torch.rand(2)
    linop = LinOpWithoutInit(mat)
    try:
        y = linop.mv(x)
        msg = ("A RuntimeError must be raised when a LinearOperator "
               "is called without .__init__() called first")
        assert False, msg
    except RuntimeError as e:
        assert "__init__" in str(e), 'The error message must contain "__init__"'

def test_linop1_shape1_err():
    mat = torch.rand(3)
    mat2 = torch.rand((3, 2))
    try:
        linop = LinOp1(mat)
        assert False, "A RuntimeError must be raised when creating a LinearOperator with 1 dimension"
    except RuntimeError:
        pass

    try:
        linop = LinOp1(mat2, is_hermitian=True)
        assert False, "A RuntimeError must be raised when creating a Hermitian LinearOperator without square shape"
    except RuntimeError:
        pass

def test_linop1_mm():
    # test if mm done correctly if not implemented
    mat = torch.rand((2, 4, 2, 3))
    x = torch.rand((3, 2))
    xv = torch.rand((3,))

    linop = LinOp1(mat)
    ymv = linop.mv(xv)
    ymm = linop.mm(x)

    assert torch.allclose(ymv, torch.matmul(mat, xv))
    assert torch.allclose(ymm, torch.matmul(mat, x))

def test_linop1_fullmatrix():
    mat = torch.rand((2, 4, 2, 3), dtype=torch.float64)
    linop = LinOp1(mat)
    linop_mat = linop.fullmatrix()
    assert torch.allclose(linop_mat, mat)

def test_linop1_rmm():
    # test if rmv and rmm done correctly if not implemented
    mat = torch.rand((2, 4, 2, 3))
    rx = torch.rand((2, 3))
    rxv = torch.rand((2,))

    linop = LinOp1(mat)
    ymv = linop.rmv(rxv)
    ymv_true = torch.matmul(mat.transpose(-2, -1), rxv)
    assert torch.allclose(ymv, ymv_true)

    ymm = linop.rmm(rx)
    ymm_true = torch.matmul(mat.transpose(-2, -1), rx)
    assert torch.allclose(ymm, ymm_true)

def test_linop2_rmm():
    # test if rmm done correctly if not implemented
    mat = torch.rand((2, 4, 2, 3))
    rx = torch.rand((2, 3))
    rxv = torch.rand((2,))

    linop = LinOp2(mat)
    ymv = linop.rmv(rxv)
    assert torch.allclose(ymv, torch.matmul(mat.transpose(-2, -1), rxv))

    ymm = linop.rmm(rx)
    assert torch.allclose(ymm, torch.matmul(mat.transpose(-2, -1), rx))

def test_linop_repr():
    dtype = torch.float32
    device = torch.device("cpu")
    mat1 = torch.randn((2, 3, 2), dtype=dtype, device=device)
    a = LinOp1(mat1)

    arepr = ["LinearOperator", "LinOp1", "(2, 3, 2)", "float32", "cpu"]
    _assert_str_contains(a.__repr__(), arepr)

    b = a.H
    brepr = arepr + ["AdjointLinearOperator", "(2, 2, 3)"]
    _assert_str_contains(b.__repr__(), brepr)

    c = b.matmul(a)
    crepr = brepr + ["MatmulLinearOperator", "(2, 2, 2)"]
    _assert_str_contains(c.__repr__(), crepr)

    d = LinearOperator.m(mat1)
    drepr = ["MatrixLinearOperator", "(2, 3, 2)"]
    _assert_str_contains(d.__repr__(), drepr)

    e = d + a
    erepr = ["AddLinearOperator", "(2, 3, 2)"]
    _assert_str_contains(d.__repr__(), drepr)

def test_linop_mat_hermit_err():
    torch.manual_seed(100)
    mat = torch.rand(3, 3)
    mat2 = torch.rand(3, 4)
    msg = "Expecting a RuntimeError for non-symmetric matrix "\
          "indicated as a Hermitian"

    try:
        LinearOperator.m(mat, is_hermitian=True)
        assert False, msg
    except RuntimeError:
        pass

    try:
        LinearOperator.m(mat2, is_hermitian=True)
        assert False, msg
    except RuntimeError:
        pass

def test_linop_op_err():
    mat1 = torch.rand(3, 4)
    vec1 = torch.rand(3)
    vec2 = torch.rand(4)
    errmsg = "Mismatch shape should raise a RuntimeError"
    linop1 = LinOp1(mat1)
    linop2 = LinOp2(mat1)

    try:
        linop1.mv(vec1)
        assert False, errmsg
    except RuntimeError:
        pass

    try:
        linop1.mm(mat1)
        assert False, errmsg
    except RuntimeError:
        pass

    try:
        linop1.rmv(vec2)
        assert False, errmsg
    except RuntimeError:
        pass

    try:
        linop1.rmm(mat1.transpose(-2, -1))
        assert False, errmsg
    except RuntimeError:
        pass

    try:
        linop1.matmul(linop2)
        assert False, errmsg
    except RuntimeError:
        pass

    try:
        linop1.H + linop2
        assert False, errmsg
    except RuntimeError:
        pass

def test_check_linop():
    mat1 = torch.rand(3, 4)
    linop1 = LinOp1(mat1)
    nlinop1 = NotLinOp1(mat1)

    with warnings.catch_warnings(record=True) as w:
        linop1.check()
        assert len(w) == 1
        assert "slow down" in str(w[0].message)

    with warnings.catch_warnings(record=True) as w:
        try:
            nlinop1.check()
            assert False, "AssertionError must be raised for non-LinearOperator"
        except AssertionError:
            pass
        assert len(w) == 1
        assert "slow down" in str(w[0].message)

def test_linop_add():
    mat = torch.randn((2, 3, 2))
    linop1 = LinOp1(mat)
    linop2 = LinOp2(mat + 1)

    # test using non-matrix linop
    c = linop1 + linop2
    assert torch.allclose(c.fullmatrix(), 2 * mat + 1)

    # test using matrix linear operator
    m1 = LinearOperator.m(mat)
    m2 = LinearOperator.m(mat + 1)
    m12 = m1 + m2
    assert torch.allclose(m12.fullmatrix(), 2 * mat + 1)

def test_linop_sub():
    # test the behaviour of subtraction of LinearOperators
    mat = torch.randn((2, 3, 2))
    linop1 = LinOp1(mat)
    linop2 = LinOp2(-mat + 1)

    # test using non-matrix linop
    c = linop1 - linop2
    assert torch.allclose(c.fullmatrix(), 2 * mat - 1)

    # test using matrix linear operator
    m1 = LinearOperator.m(mat)
    m2 = LinearOperator.m(-mat + 1)
    m12 = m1 - m2
    assert torch.allclose(m12.fullmatrix(), 2 * mat - 1)

def test_linop_mul():
    # test the behaviour of multiplication of LinearOperator with a number
    mat = torch.randn((2, 3, 2))
    linop1 = LinOp1(mat)
    linop2 = LinearOperator.m(mat)

    for f1 in [2, 4.0]:
        print(f1)
        # test using non-matrix linop multiplier
        c11l = linop1 * f1
        c11r = f1 * linop1
        # test using matrix linop multiplier
        c12l = linop2 * f1
        c12r = f1 * linop2
        assert torch.allclose(c11l.fullmatrix(), f1 * mat)
        assert torch.allclose(c11r.fullmatrix(), f1 * mat)
        assert torch.allclose(c12l.fullmatrix(), f1 * mat)
        assert torch.allclose(c12r.fullmatrix(), f1 * mat)

def _assert_str_contains(s, slist):
    for c in slist:
        assert c in s

############# special case tests #############

# #1 case: various cases uses linop in editable module
class ClassWithLinop(EditableModule):
    def __init__(self):
        mat = torch.randn(3, 3)
        mat = (mat + mat.transpose(-2, -1)) * 0.5
        # using LinOp1 instead of LinearOperator.m to detach .fullmatrix()
        # output with mat
        self.linop = LinOp1(mat, is_hermitian=True)

    def symeig1(self, method):
        return symeig(self.linop, method=method)[1]

    def getparamnames(self, methodname, prefix=""):
        if methodname == "symeig1":
            return self.linop._getparamnames(prefix=prefix + "linop.")
        else:
            raise KeyError()

def test_assertparams_if_fullmatrix_is_called_1():
    a = ClassWithLinop()
    # if fullmatrix stores cache, then assertparams will fail because the
    # method a.symeig1 changes the object's states
    a.assertparams(a.symeig1, "exacteig")

def test_assertparams_if_fullmatrix_is_called_2():
    # This test is involving the linear algebra method that requires the
    # fullmatrix.
    # If ClassWithLinop.symeig2 is called, the method depends on the full matrix
    # of the linop, so if fullmatrix() cache is stored, then it does
    # not directly depend on parameters of linop (but related to the
    # fullmatrix of linop), which makes assertparams raises warnings

    a = ClassWithLinop()

    # calling symeig1 first
    a.symeig1(method="exacteig")
    # doing assert params to make sure there is no cache is stored in
    # fullmatrix (in the old version, this failed)
    with warnings.catch_warnings(record=True) as w:
        a.assertparams(a.symeig1, "exacteig")
        assert len(w) == 0



from typing import Callable, Mapping, Any, Sequence, Union, List
import torch
from xitorch._utils.misc import TensorNonTensorSeparator, get_method
from xitorch._utils.assertfuncs import assert_fcn_params
from xitorch._impls.optimize.root.rootsolver import broyden1, broyden2, \
    linearmixing
from xitorch._impls.optimize.equilibrium import anderson_acc
from xitorch._impls.optimize.minimizer import gd, adam
from xitorch.linalg.solve import solve
from xitorch.grad.jachess import jac
from xitorch._core.pure_function import get_pure_function, make_sibling
from xitorch._docstr.api_docstr import get_methods_docstr
from xitorch.debug.modes import is_debug_enabled

__all__ = ["equilibrium", "rootfinder", "minimize"]

_RF_METHODS = {
    "broyden1": broyden1,
    "broyden2": broyden2,
    "linearmixing": linearmixing,
}

_EQUIL_METHODS = {
    # equilibrium can use all rootfinder methods, but there are several methods developed specifically for
    # equilibrium (or fixed-point iterations). This dictionary gives the list of those special methods.
    "anderson_acc": anderson_acc,
}

_OPT_METHODS = {
    "gd": gd,
    "adam": adam,
}

def rootfinder(
        fcn: Callable[..., torch.Tensor],
        y0: torch.Tensor,
        params: Sequence[Any] = [],
        bck_options: Mapping[str, Any] = {},
        method: Union[str, Callable, None] = None,
        **fwd_options) -> torch.Tensor:
    r"""
    Solving the rootfinder equation of a given function,

    .. math::

        \mathbf{0} = \mathbf{f}(\mathbf{y}, \theta)

    where :math:`\mathbf{f}` is a function that can be non-linear and
    produce output of the same shape of :math:`\mathbf{y}`, and :math:`\theta`
    is other parameters required in the function.
    The output of this block is :math:`\mathbf{y}`
    that produces the :math:`\mathbf{0}` as the output.

    Arguments
    ---------
    fcn : callable
        The function :math:`\mathbf{f}` with output tensor ``(*ny)``
    y0 : torch.tensor
        Initial guess of the solution with shape ``(*ny)``
    params : list
        Sequence of any other parameters to be put in ``fcn``
    bck_options : dict
        Method-specific options for the backward solve (see :func:`xitorch.linalg.solve`)
    method : str or callable or None
        Rootfinder method. If None, it will choose ``"broyden1"``.
    **fwd_options
        Method-specific options (see method section)

    Returns
    -------
    torch.tensor
        The solution which satisfies
        :math:`\mathbf{0} = \mathbf{f}(\mathbf{y},\theta)`
        with shape ``(*ny)``

    Example
    -------
    .. testsetup:: root1

        import torch
        from xitorch.optimize import rootfinder

    .. doctest:: root1

        >>> def func1(y, A):  # example function
        ...     return torch.tanh(A @ y + 0.1) + y / 2.0
        >>> A = torch.tensor([[1.1, 0.4], [0.3, 0.8]]).requires_grad_()
        >>> y0 = torch.zeros((2,1))  # zeros as the initial guess
        >>> yroot = rootfinder(func1, y0, params=(A,))
        >>> print(yroot)
        tensor([[-0.0459],
                [-0.0663]], grad_fn=<_RootFinderBackward>)
    """
    # perform implementation check if debug mode is enabled
    if is_debug_enabled():
        assert_fcn_params(fcn, (y0, *params))

    pfunc = get_pure_function(fcn)
    fwd_options["method"] = _get_rootfinder_default_method(method)
    return _RootFinder.apply(pfunc, y0, pfunc, "rootfinder", fwd_options, bck_options,
                             len(params), *params, *pfunc.objparams())

def equilibrium(
        fcn: Callable[..., torch.Tensor],
        y0: torch.Tensor,
        params: Sequence[Any] = [],
        bck_options: Mapping[str, Any] = {},
        method: Union[str, Callable, None] = None,
        **fwd_options) -> torch.Tensor:
    r"""
    Solving the equilibrium equation of a given function,

    .. math::

        \mathbf{y} = \mathbf{f}(\mathbf{y}, \theta)

    where :math:`\mathbf{f}` is a function that can be non-linear and
    produce output of the same shape of :math:`\mathbf{y}`, and :math:`\theta`
    is other parameters required in the function.
    The output of this block is :math:`\mathbf{y}`
    that produces the same :math:`\mathbf{y}` as the output.

    Arguments
    ---------
    fcn : callable
        The function :math:`\mathbf{f}` with output tensor ``(*ny)``
    y0 : torch.tensor
        Initial guess of the solution with shape ``(*ny)``
    params : list
        Sequence of any other parameters to be put in ``fcn``
    bck_options : dict
        Method-specific options for the backward solve (see :func:`xitorch.linalg.solve`)
    method : str or None
        Rootfinder method. If None, it will choose ``"broyden1"``.
    **fwd_options
        Method-specific options (see method section)

    Returns
    -------
    torch.tensor
        The solution which satisfies
        :math:`\mathbf{y} = \mathbf{f}(\mathbf{y},\theta)`
        with shape ``(*ny)``

    Example
    -------
    .. testsetup:: equil1

        import torch
        from xitorch.optimize import equilibrium

    .. doctest:: equil1

        >>> def func1(y, A):  # example function
        ...     return torch.tanh(A @ y + 0.1) + y / 2.0
        >>> A = torch.tensor([[1.1, 0.4], [0.3, 0.8]]).requires_grad_()
        >>> y0 = torch.zeros((2,1))  # zeros as the initial guess
        >>> yequil = equilibrium(func1, y0, params=(A,))
        >>> print(yequil)
        tensor([[ 0.2313],
                [-0.5957]], grad_fn=<_RootFinderBackward>)

    Note
    ----
    * This is a direct implementation of finding the root of
      :math:`\mathbf{g}(\mathbf{y}, \theta) = \mathbf{y} - \mathbf{f}(\mathbf{y}, \theta)`
    """
    # perform implementation check if debug mode is enabled
    if is_debug_enabled():
        assert_fcn_params(fcn, (y0, *params))

    pfunc = get_pure_function(fcn)

    @make_sibling(pfunc)
    def new_fcn(y, *params):
        return y - pfunc(y, *params)

    method = _get_equilibrium_default_method(method)
    fwd_options["method"] = method
    fwd_fcn = pfunc if method in _EQUIL_METHODS else new_fcn
    alg_type = "equilibrium" if method in _EQUIL_METHODS else "rootfinder"
    return _RootFinder.apply(new_fcn, y0, fwd_fcn, alg_type, fwd_options, bck_options,
                             len(params), *params, *pfunc.objparams())

def minimize(
        fcn: Callable[..., torch.Tensor],
        y0: torch.Tensor,
        params: Sequence[Any] = [],
        bck_options: Mapping[str, Any] = {},
        method: Union[None, str, Callable] = None,
        **fwd_options) -> torch.Tensor:
    r"""
    Solve the unbounded minimization problem:

    .. math::

        \mathbf{y^*} = \arg\min_\mathbf{y} f(\mathbf{y}, \theta)

    to find the best :math:`\mathbf{y}` that minimizes the output of the
    function :math:`f`.

    Arguments
    ---------
    fcn: callable
        The function to be optimized with output tensor with 1 element.
    y0: torch.tensor
        Initial guess of the solution with shape ``(*ny)``
    params: list
        Sequence of any other parameters to be put in ``fcn``
    bck_options: dict
        Method-specific options for the backward solve (see :func:`xitorch.linalg.solve`)
    method: str or callable or None
        Minimization method. If None, it will choose ``"broyden1"``.
    **fwd_options
        Method-specific options (see method section)

    Returns
    -------
    torch.tensor
        The solution of the minimization with shape ``(*ny)``

    Example
    -------
    .. testsetup:: root1

        import torch
        from xitorch.optimize import minimize

    .. doctest:: root1

        >>> def func1(y, A):  # example function
        ...     return torch.sum((A @ y)**2 + y / 2.0)
        >>> A = torch.tensor([[1.1, 0.4], [0.3, 0.8]]).requires_grad_()
        >>> y0 = torch.zeros((2,1))  # zeros as the initial guess
        >>> ymin = minimize(func1, y0, params=(A,))
        >>> print(ymin)
        tensor([[-0.0519],
                [-0.2684]], grad_fn=<_RootFinderBackward>)
    """

    assert not torch.is_complex(y0), \
        "complex number is not supported on xitorch.optimize.rootfinder at the moment"

    # perform implementation check if debug mode is enabled
    if is_debug_enabled():
        assert_fcn_params(fcn, (y0, *params))

    pfunc = get_pure_function(fcn)

    fwd_options["method"] = _get_minimizer_default_method(method)
    method = fwd_options["method"]

    # minimization can use rootfinder algorithm, so check if it is actually
    # using the optimization algorithm, not the rootfinder algorithm
    opt_method = method not in _RF_METHODS.keys()

    # the rootfinder algorithms are designed to move to the opposite direction
    # of the output of the function, so the output of this function is just
    # the grad of z w.r.t. y
    # if it is going to optimization method, then also returns the value

    @make_sibling(pfunc)
    def _min_fwd_fcn(y, *params):
        with torch.enable_grad():
            y1 = y.clone().requires_grad_()
            z = pfunc(y1, *params)
        grady, = torch.autograd.grad(z, (y1,), retain_graph=True,
                                     create_graph=torch.is_grad_enabled())
        return z, grady

    @make_sibling(_min_fwd_fcn)
    def _rf_fcn(y, *params):
        z, grady = _min_fwd_fcn(y, *params)
        return grady

    # if using the optimization algorithm, then the forward function is the one
    # that returns f and grad
    if opt_method:
        _fwd_fcn = _min_fwd_fcn
    # if it is just using the rootfinder algorithm, then the forward function
    # is the one that returns only the grad
    else:
        _fwd_fcn = _rf_fcn

    alg_type = "minimizer" if opt_method else "rootfinder"
    return _RootFinder.apply(_rf_fcn, y0, _fwd_fcn, alg_type, fwd_options, bck_options,
                             len(params), *params, *pfunc.objparams())

class _RootFinder(torch.autograd.Function):
    @staticmethod
    def forward(ctx, fcn, y0, fwd_fcn, alg_type, options, bck_options, nparams, *allparams):
        # fcn: a function that returns what has to be 0 (will be used in the
        #      backward, not used in the forward). For minimization, it is
        #      the gradient
        # fwd_fcn: a function that will be executed in the forward method
        #          (unused in the backward)
        # This class is also used for minimization, where fcn and fwd_fcn might
        # be slightly different

        # set default options
        config = options
        ctx.bck_options = bck_options

        params = allparams[:nparams]
        objparams = allparams[nparams:]

        with fwd_fcn.useobjparams(objparams):

            method = config.pop("method")
            methods = {
                "minimizer": _OPT_METHODS,
                "rootfinder": _RF_METHODS,
                "equilibrium": _EQUIL_METHODS
            }[alg_type]
            name = alg_type
            method_fcn = get_method(name, methods, method)
            y = method_fcn(fwd_fcn, y0, params, **config)

        ctx.fcn = fcn

        # split tensors and non-tensors params
        ctx.nparams = nparams
        ctx.param_sep = TensorNonTensorSeparator(allparams)
        tensor_params = ctx.param_sep.get_tensor_params()
        ctx.save_for_backward(y, *tensor_params)

        return y

    @staticmethod
    def backward(ctx, grad_yout):
        param_sep = ctx.param_sep
        yout = ctx.saved_tensors[0]
        nparams = ctx.nparams
        fcn = ctx.fcn

        # merge the tensor and nontensor parameters
        tensor_params = ctx.saved_tensors[1:]
        allparams = param_sep.reconstruct_params(tensor_params)
        params = allparams[:nparams]
        objparams = allparams[nparams:]

        # dL/df
        with ctx.fcn.useobjparams(objparams):

            jac_dfdy = jac(fcn, params=(yout, *params), idxs=[0])[0]
            gyfcn = solve(A=jac_dfdy.H, B=-grad_yout.reshape(-1, 1),
                          bck_options=ctx.bck_options, **ctx.bck_options)
            gyfcn = gyfcn.reshape(grad_yout.shape)

            # get the grad for the params
            with torch.enable_grad():
                tensor_params_copy = [p.clone().requires_grad_() for p in tensor_params]
                allparams_copy = param_sep.reconstruct_params(tensor_params_copy)
                params_copy = allparams_copy[:nparams]
                objparams_copy = allparams_copy[nparams:]
                with ctx.fcn.useobjparams(objparams_copy):
                    yfcn = fcn(yout, *params_copy)

            grad_tensor_params = torch.autograd.grad(yfcn, tensor_params_copy, grad_outputs=gyfcn,
                                                     create_graph=torch.is_grad_enabled(),
                                                     allow_unused=True)
            grad_nontensor_params = [None for _ in range(param_sep.nnontensors())]
            grad_params = param_sep.reconstruct_params(grad_tensor_params, grad_nontensor_params)

        return (None, None, None, None, None, None, None, *grad_params)

def _get_rootfinder_default_method(method):
    if method is None:
        return "broyden1"
    else:
        return method

def _get_equilibrium_default_method(method):
    if method is None:
        return _get_rootfinder_default_method(method)
    else:
        return method

def _get_minimizer_default_method(method):
    if method is None:
        return "broyden1"
    else:
        return method


# docstring completion
rf_methods: List[Callable] = [broyden1, broyden2, linearmixing]
eq_methods: List[Callable] = rf_methods + [anderson_acc]
opt_methods: List[Callable] = [gd, adam]
rootfinder.__doc__ = get_methods_docstr(rootfinder, rf_methods)
equilibrium.__doc__ = get_methods_docstr(equilibrium, eq_methods)
minimize.__doc__ = get_methods_docstr(minimize, rf_methods + opt_methods)



import numpy as np
import torch

# no gradient flowing in the following functions

def leggauss(fcn, xl, xu, params, n=100, **unused):
    """
    Performing 1D integration using Legendre-Gaussian quadrature

    Keyword arguments
    -----------------
    n: int
        The number of integration points.
    """
    xlg, wlg = np.polynomial.legendre.leggauss(n)
    ndim = len(xu.shape)
    xlg = torch.tensor(xlg, dtype=xu.dtype, device=xu.device)[(...,) + (None,) * ndim]  # (n, *nx)
    wlg = torch.tensor(wlg, dtype=xu.dtype, device=xu.device)[(...,) + (None,) * ndim]  # (n, *nx)
    wlg *= 0.5 * (xu - xl)
    xs = xlg * (0.5 * (xu - xl)) + (0.5 * (xu + xl))  # (n, *nx)

    res = wlg[0] * fcn(xs[0], *params)
    for i in range(1, n):
        res += wlg[i] * fcn(xs[i], *params)
    return res



from typing import Tuple
import torch

def normalize_bcast_dims(*shapes):
    """
    Normalize the lengths of the input shapes to have the same length.
    The shapes are padded at the front by 1 to make the lengths equal.
    """
    maxlens = max([len(shape) for shape in shapes])
    res = [[1] * (maxlens - len(shape)) + list(shape) for shape in shapes]
    return res

def get_bcasted_dims(*shapes):
    """
    Return the broadcasted shape of the given shapes.
    """
    shapes = normalize_bcast_dims(*shapes)
    return [max(*a) for a in zip(*shapes)]

def match_dim(*xs: torch.Tensor, contiguous: bool = False) -> Tuple[torch.Tensor, ...]:
    # match the N-1 dimensions of x and xq for searchsorted and gather with dim=-1
    orig_shapes = tuple(x.shape[:-1] for x in xs)
    shape = tuple(get_bcasted_dims(*orig_shapes))
    xs_new = tuple(x.expand(shape + (x.shape[-1],)) for x in xs)
    if contiguous:
        xs_new = tuple(x.contiguous() for x in xs_new)
    return xs_new



import torch
import numpy as np

__all__ = ["get_np_dtype"]

def get_np_dtype(dtype: torch.dtype):
    # return the corresponding numpy dtype from the input pytorch's tensor dtype
    if dtype == torch.float32:
        return np.float32
    elif dtype == torch.float64:
        return np.float64
    elif dtype == torch.complex64:
        return np.complex64
    elif dtype == torch.complex128:
        return np.complex128
    else:
        raise TypeError("Unknown type: %s" % dtype)



from __future__ import annotations
from typing import Any, Optional, List, Tuple, Dict
from copy import deepcopy, copy
import torch

__all__ = ["Packer"]

class Packer(object):
    """
    Packer is an object that could extract the tensors in a structure and
    rebuild the structure from the given tensors.
    This object preserves the structure of the object by performing the deepcopy
    of the object, except for the tensor.

    Arguments
    ---------
    obj: Any
        Any structure object that contains tensors.

    Example
    -------
    .. testsetup::

        import torch
        import xitorch

    .. doctest::

        >>> a = torch.tensor(1.0)
        >>> obj = {
        ...     "a": a,
        ...     "b": a * 3,
        ...     "c": a,
        ... }
        >>> packer = xitorch.Packer(obj)
        >>> tensors = packer.get_param_tensor_list()
        >>> print(tensors)
        [tensor(1.), tensor(3.)]
        >>> new_tensors = [torch.tensor(2.0), torch.tensor(4.0)]
        >>> new_obj = packer.construct_from_tensor_list(new_tensors)
        >>> print(new_obj)
        {'a': tensor(2.), 'b': tensor(4.), 'c': tensor(2.)}
    """
    def __init__(self, obj: Any):
        # deep copy the object, except the tensors
        tensor_lists = _extract_tensors(obj)
        memo = {id(t): t for t in tensor_lists}
        self._tensor_memo = copy(memo)  # shallow copy
        self._obj = deepcopy(obj, memo)

        # caches
        self._params_tensor_list: Optional[List[torch.Tensor]] = tensor_lists
        self._unique_params_idxs: Optional[List[int]] = None
        self._unique_inverse_idxs: Optional[List[int]] = None
        self._unique_tensor_shapes: Optional[List[torch.Size]] = None
        self._tensor_shapes: Optional[List[torch.Size]] = None
        self._unique_tensor_numels: Optional[List[int]] = None
        self._unique_tensor_numel_tot: Optional[int] = None
        self._tensor_numels: Optional[List[int]] = None
        self._tensor_numel_tot: Optional[int] = None

    def get_param_tensor_list(self, unique: bool = True) -> List[torch.Tensor]:
        """
        Returns the list of tensors contained in the object. It will traverse
        down the object via elements for list, values for dictionary, or
        ``__dict__`` for object that has ``__dict__`` attribute.

        Arguments
        ---------
        unique: bool
            If True, then only returns the unique tensors. Otherwise, duplicates
            can also be returned.

        Returns
        -------
        list of torch.Tensor
            List of tensors contained in the object.
        """
        # get the params tensor list
        if self._params_tensor_list is not None:
            # get it from cache if available
            params_tensors = self._params_tensor_list
        else:
            params_tensors = _extract_tensors(self._obj)
            self._params_tensor_list = params_tensors

        # only take the unique tensors if required
        if unique:
            if self._unique_params_idxs is not None:
                unique_idxs = self._unique_params_idxs
                unique_inverse = self._unique_inverse_idxs
            else:
                unique_idxs, unique_inverse = _get_unique_idxs(params_tensors)
                self._unique_params_idxs = unique_idxs
                self._unique_inverse_idxs = unique_inverse

            params_tensors = [params_tensors[i] for i in unique_idxs]

        if unique:
            self._unique_tensor_shapes = [p.shape for p in params_tensors]
        else:
            self._tensor_shapes = [p.shape for p in params_tensors]

        return params_tensors

    def get_param_tensor(self, unique: bool = True) -> Optional[torch.Tensor]:
        """
        Returns the tensor parameters as a single tensor. This can be used,
        for example, if there are multiple parameters to be optimized using
        ``xitorch.optimize.minimize``.

        Arguments
        ---------
        unique: bool
            If True, then only returns the tensor from unique tensors list.
            Otherwise, duplicates can also be returned.

        Returns
        -------
        torch.Tensor or None
            The parameters of the object in a single tensor or None if there
            is no tensor contained in the object.
        """
        params = self.get_param_tensor_list(unique=unique)
        if len(params) == 0:
            return None
        else:
            if unique:
                self._unique_tensor_numels = [p.numel() for p in params]
                self._unique_tensor_numel_tot = sum(self._unique_tensor_numels)
            else:
                self._tensor_numels = [p.numel() for p in params]
                self._tensor_numel_tot = sum(self._tensor_numels)
            if len(params) == 1:
                return params[0]
            else:
                tparam = torch.cat([p.reshape(-1) for p in params])
                return tparam

    def construct_from_tensor_list(self, tensors: List[torch.Tensor], unique: bool = True) -> Any:
        """
        Construct the object from the tensor list and returns the object structure
        with the new tensors. Executing this does not change the state of the
        Packer object.

        Arguments
        ---------
        tensors: list of torch.Tensor
            The tensor parameters to be filled into the object.
        unique: bool
            Indicating if the tensor list ``tensors`` is from the unique
            parameters of the object.

        Returns
        -------
        Any
            A new object with the same structure as the input to ``__init__``
            object except the tensor is changed according to ``tensors``.
        """
        if unique:
            tensor_shapes = self._unique_tensor_shapes
        else:
            tensor_shapes = self._tensor_shapes

        if tensor_shapes is None:
            raise RuntimeError("Please execute self.get_param_tensor_list(%s) first" % str(unique))
        else:
            # make sure the length matches
            if len(tensor_shapes) != len(tensors):
                raise RuntimeError("Mismatch length of the tensors")

            if len(tensor_shapes) == 0:
                return self._obj

            # check the tensor shapes
            for i, (tens, shape) in enumerate(zip(tensors, tensor_shapes)):
                if tens.shape != shape:
                    msg = "The tensors[%d] has mismatch shape from the original. Expected: %s, got: %s" % \
                          (i, tens.shape, shape)
                    raise RuntimeError(msg)

            # duplicate the tensors if the input is unique list of tensors
            if unique:
                assert self._unique_inverse_idxs, "Please report to Github"
                tensors = [tensors[self._unique_inverse_idxs[i]] for i in range(len(self._unique_inverse_idxs))]
            else:
                # _put_tensors will change the tensors, so this is just to preserve
                # the input
                tensors = copy(tensors)

            # deepcopy the object, except the tensors
            memo = copy(self._tensor_memo)
            new_obj = deepcopy(self._obj, memo)
            new_obj = _put_tensors(new_obj, tensors)

            return new_obj

    def construct_from_tensor(self, a: torch.Tensor, unique: bool = True) -> Any:
        """
        Construct the object from the single tensor (i.e. it is the parameters
        tensor merged into a single tensor) and returns the object structure
        with the new tensor. Executing this does not change the state of the
        Packer object.

        Arguments
        ---------
        a: torch.Tensor
            The single tensor parameter to be filled.
        unique: bool
            Indicating if the tensor ``a`` is from the unique parameters of the
            object.

        Returns
        -------
        Any
            A new object with the same structure as the input to ``__init__``
            object except the tensor is changed according to ``a``.
        """
        if unique:
            tensor_shapes = self._unique_tensor_shapes
            tensor_numel_tot = self._unique_tensor_numel_tot
            tensor_numels = self._unique_tensor_numels
        else:
            tensor_shapes = self._tensor_shapes
            tensor_numel_tot = self._tensor_numel_tot
            tensor_numels = self._tensor_numels

        if tensor_shapes is None:
            raise RuntimeError("Please execute self.get_param_tensor(%s) first" % str(unique))
        elif len(tensor_shapes) == 0:
            return self._obj
        else:
            assert tensor_numel_tot is not None, "Please report to Github"
            assert tensor_numels is not None, "Please report to Github"
            if a.numel() != tensor_numel_tot:
                msg = "The number of element does not match. Expected: %d, got: %d" % \
                      (tensor_numel_tot, a.numel())
                raise RuntimeError(msg)

            if len(tensor_numels) == 1:
                params: List[torch.Tensor] = [a]
            else:
                # reshape the parameters
                ioffset = 0
                params = []
                for i in range(len(tensor_numels)):
                    p = a[ioffset:ioffset + tensor_numels[i]].reshape(tensor_shapes[i])
                    ioffset += tensor_numels[i]
                    params.append(p)

            return self.construct_from_tensor_list(params, unique=unique)

def _extract_tensors(b: Any) -> List[torch.Tensor]:
    # extract all the tensors from the given object
    # this function traverses down the object to collect all the tensors

    res: List[torch.Tensor] = []
    if isinstance(b, torch.Tensor):
        res.append(b)
    elif isinstance(b, list):
        for elmt in b:
            res.extend(_extract_tensors(elmt))
    elif isinstance(b, dict):
        for elmt in b.values():
            res.extend(_extract_tensors(elmt))
    elif hasattr(b, "__dict__"):
        for elmt in b.__dict__.values():
            res.extend(_extract_tensors(elmt))
    return res

def _put_tensors(b: Any, tensors: List) -> Any:
    # put the tensors recursively in the object, with the same order as
    # _extract_tensors.
    # the tensors will be changed in this class, so make sure to have
    # a shallow copy if you want to preserve your input
    if isinstance(b, torch.Tensor):
        b = tensors.pop(0)
    elif isinstance(b, list):
        for i, elmt in enumerate(b):
            b[i] = _put_tensors(elmt, tensors)
    elif isinstance(b, dict):
        for key, elmt in b.items():
            b[key] = _put_tensors(elmt, tensors)
    elif hasattr(b, "__dict__"):
        for key, elmt in b.__dict__.items():
            b.__dict__[key] = _put_tensors(elmt, tensors)
    return b

def _get_unique_idxs(b: List) -> Tuple[List[int], List[int]]:
    # get unique indices based on the ids of the b's elements
    # and the index for inversing the unique process

    ids_list = [id(bb) for bb in b]
    unique_ids: Dict[int, int] = {}
    unique_idxs: List[int] = []
    unique_inverse: List[int] = []
    for i, idnum in enumerate(ids_list):
        if idnum in unique_ids:
            unique_inverse.append(unique_ids[idnum])
        else:
            unique_ids[idnum] = len(unique_idxs)
            unique_idxs.append(i)
            unique_inverse.append(unique_ids[idnum])
    return unique_idxs, unique_inverse



import sys
import argparse
from xitorch.debug.modes import enable_debug

def parse_args():
    # parse the argument
    parser = argparse.ArgumentParser("Run python script by enabling xitorch debug mode")
    parser.add_argument("scriptfile", type=str, help="Path to the script to run")
    parser.add_argument("args", type=str, nargs=argparse.REMAINDER,
                        help="The arguments needed to run the script")
    return parser.parse_args()

def main():
    args = parse_args()

    scriptfile = args.scriptfile
    scriptargs = args.args if args.args is not None else []
    scriptargs.insert(0, scriptfile)
    sys.argv[:] = scriptargs[:]

    # compile and run the code with debug mode enabled
    with enable_debug():
        with open(scriptfile, 'rb') as stream:
            code = compile(stream.read(), scriptfile, 'exec')
        globs = {
            '__file__': scriptfile,
            '__name__': '__main__',
            '__package__': None,
            '__cached__': None,
        }
        exec(code, globs, None)


if __name__ == "__main__":
    main()



from typing import Optional, List, Union, Callable
import torch
from xitorch._core.editable_module import EditableModule
from xitorch._impls.interpolate.interp_1d import CubicSpline1D, LinearInterp1D
from xitorch._docstr.api_docstr import get_methods_docstr
from xitorch._utils.bcast import match_dim
from xitorch._utils.misc import get_method

__all__ = ["Interp1D"]

class Interp1D(EditableModule):
    """
    1D interpolation class. When initializing the class, the `x` must be
    specified and `y` can be specified during initialization or later.

    Arguments
    ---------
    x: torch.Tensor
        The position of known values in tensor with shape ``(..., nr)``
    y: torch.Tensor or None
        The values at the given position with shape ``(..., nr)``.
        If ``None``, it must be supplied during ``__call__``
    method: str or callable or None
        Interpolation method. If None, it will choose ``"cspline"``.
    assume_sorted: bool
        Assume x is sorted monotonically increasing. If False, then it sorts the
        input x and y first before doing the interpolation.
    **fwd_options
        Method-specific options (see method section below)

    Note
    ----
    Batched ``x`` and ``xq`` is only implemented if there is no extrapolation involved.
    """

    def __init__(self,
                 x: torch.Tensor,
                 y: Optional[torch.Tensor] = None,
                 method: Union[str, Callable, None] = None,
                 assume_sorted: bool = False,
                 **fwd_options):
        if method is None:
            method = "cspline"
        methods = {
            "cspline": CubicSpline1D,
            "linear": LinearInterp1D,
        }
        method_cls = get_method("Interp1D", methods, method)

        # sort x
        self.idx: Optional[torch.Tensor] = None
        if not assume_sorted:
            x, idx = torch.sort(x, dim=-1)
            if y is not None:
                y, idx = match_dim(y, idx)
                y = torch.gather(y, dim=-1, index=idx)
            else:
                self.idx = idx

        self.obj = method_cls(x, y, **fwd_options)

    def __call__(self, xq: torch.Tensor, y: Optional[torch.Tensor] = None) -> torch.Tensor:
        """
        Arguments
        ----------------
        xq: torch.Tensor
            The position of query points with shape ``(..., nrq)``.
        y: torch.Tensor or None
            The values at the given position with shape ``(..., nr)``.
            If ``y`` has been specified during ``__init__`` and also
            specified here, the value of ``y`` given here will be ignored.
            If no ``y`` ever specified, then it will raise an error.

        Returns
        -------
        torch.Tensor
            The interpolated values with shape ``(..., nrq)``.
        """
        if self.idx is not None and y is not None:
            y, idx = match_dim(y, self.idx)
            y = torch.gather(y, dim=-1, index=idx)
        return self.obj(xq, y)

    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """"""
        return [prefix + "obj." + c for c in self.obj.getparamnames()]


# docstring completion
interp1d_methods = {
    "cspline": CubicSpline1D,
    "linear": LinearInterp1D,
}
Interp1D.__doc__ = get_methods_docstr(Interp1D, interp1d_methods)



import re
import ast

__all__ = ["get_attr", "set_attr", "del_attr"]

# pattern to split the names, e.g. "model.params[1]" into ["model", "params", "[1]"]
sp = re.compile(r"\[{0,1}[\"']{0,1}\w+[\"']{0,1}\]{0,1}")

def get_attr(obj, name):
    return _get_attr(obj, _preproc_name(name))

def set_attr(obj, name, val):
    return _set_attr(obj, _preproc_name(name), val)

def del_attr(obj, name):
    return _del_attr(obj, _preproc_name(name))


def _get_attr(obj, names):
    attrfcn = lambda obj, name: getattr(obj, name)
    dictfcn = lambda obj, key: obj.__getitem__(key)
    listfcn = lambda obj, key: obj.__getitem__(key)
    return _traverse_attr(obj, names, attrfcn, dictfcn, listfcn)

def _set_attr(obj, names, val):
    attrfcn = lambda obj, name: setattr(obj, name, val)
    dictfcn = lambda obj, key: obj.__setitem__(key, val)
    listfcn = lambda obj, key: obj.__setitem__(key, val)
    return _traverse_attr(obj, names, attrfcn, dictfcn, listfcn)

def _del_attr(obj, names):
    attrfcn = lambda obj, name: delattr(obj, name)
    dictfcn = lambda obj, key: obj.__delitem__(key)

    def listfcn(obj, key):
        obj.__delitem__(key)
        obj.insert(key, None)  # to preserve the length
    return _traverse_attr(obj, names, attrfcn, dictfcn, listfcn)


def _preproc_name(name):
    return sp.findall(name)

def _traverse_attr(obj, names, attrfcn, dictfcn, listfcn):
    if len(names) == 1:
        return _applyfcn(obj, names[0], attrfcn, dictfcn, listfcn)
    else:
        return _applyfcn(_get_attr(obj, names[:-1]), names[-1], attrfcn, dictfcn, listfcn)

def _applyfcn(obj, name, attrfcn, dictfcn, listfcn):
    if name[0] == "[":
        key = ast.literal_eval(name[1:-1])
        if isinstance(obj, dict):
            return dictfcn(obj, key)
        elif isinstance(obj, list):
            return listfcn(obj, key)
        else:
            msg = "The parameter with [] must be either a dictionary or a list. "
            msg += "Got type: %s" % type(obj)
            raise TypeError(msg)
    else:
        return attrfcn(obj, name)



import warnings
import torch
import pytest
from torch.autograd import gradcheck, gradgradcheck
from xitorch.debug.modes import enable_debug
from xitorch import LinearOperator
from xitorch.linalg.symeig import lsymeig, symeig, svd
from xitorch.linalg.solve import solve
from xitorch._utils.bcast import get_bcasted_dims
from xitorch._utils.exceptions import MathWarning
from xitorch._tests.utils import device_dtype_float_test

seed = 12345

############## lsymeig ##############
@device_dtype_float_test()
def test_lsymeig_nonhermit_err(dtype, device):
    torch.manual_seed(seed)
    mat = torch.rand((3, 3), dtype=dtype, device=device)
    linop = LinearOperator.m(mat, False)
    linop2 = LinearOperator.m(mat + mat.transpose(-2, -1), True)

    try:
        res = lsymeig(linop)
        assert False, "A RuntimeError must be raised if the A linear operator in lsymeig is not Hermitian"
    except RuntimeError:
        pass

    try:
        res = lsymeig(linop2, M=linop)
        assert False, "A RuntimeError must be raised if the M linear operator in lsymeig is not Hermitian"
    except RuntimeError:
        pass

@device_dtype_float_test()
def test_lsymeig_mismatch_err(dtype, device):
    torch.manual_seed(seed)
    mat1 = torch.rand((3, 3), dtype=dtype, device=device)
    mat2 = torch.rand((2, 2), dtype=dtype, device=device)
    mat1 = mat1 + mat1.transpose(-2, -1)
    mat2 = mat2 + mat2.transpose(-2, -1)
    linop1 = LinearOperator.m(mat1, True)
    linop2 = LinearOperator.m(mat2, True)

    try:
        res = lsymeig(linop1, M=linop2)
        assert False, "A RuntimeError must be raised if A & M shape are mismatch"
    except RuntimeError:
        pass


@device_dtype_float_test(only64=True, include_complex=True, additional_kwargs={
    "shape": [(4, 4), (2, 4, 4), (2, 3, 4, 4)],
    "method": ["exacteig", "custom_exacteig"],  # only 2 of methods, because both gradient implementations are covered
})
def test_lsymeig_A(dtype, device, shape, method):
    torch.manual_seed(seed)
    mat1 = torch.rand(shape, dtype=dtype, device=device)
    mat1 = mat1 + mat1.transpose(-2, -1).conj()
    mat1 = mat1.requires_grad_()
    linop1 = LinearOperator.m(mat1, True)
    fwd_options = {"method": method}

    for neig in [2, shape[-1]]:
        eigvals, eigvecs = lsymeig(linop1, neig=neig, **fwd_options)  # eigvals: (..., neig), eigvecs: (..., na, neig)
        eigvals = eigvals.to(eigvecs.dtype)
        assert list(eigvecs.shape) == list([*linop1.shape[:-1], neig])
        assert list(eigvals.shape) == list([*linop1.shape[:-2], neig])

        ax = linop1.mm(eigvecs)
        xe = torch.matmul(eigvecs, torch.diag_embed(eigvals, dim1=-2, dim2=-1))
        assert torch.allclose(ax, xe)

        # only perform gradcheck if neig is full, to reduce the computational cost
        if neig == shape[-1]:
            def lsymeig_fcn(amat):
                amat = (amat + amat.transpose(-2, -1).conj()) * 0.5  # symmetrize
                alinop = LinearOperator.m(amat, is_hermitian=True)
                eigvals_, eigvecs_ = lsymeig(alinop, neig=neig, **fwd_options)
                return eigvals_, eigvecs_.abs()

            gradcheck(lsymeig_fcn, (mat1,))
            gradgradcheck(lsymeig_fcn, (mat1,))

@device_dtype_float_test(only64=True, include_complex=True, additional_kwargs={
    "ashape": [(3, 3), (2, 3, 3), (2, 1, 3, 3)],
    "mshape": [(3, 3), (2, 3, 3), (2, 1, 3, 3)],
    "method": ["exacteig", "custom_exacteig"],  # only 2 of methods, because both gradient implementations are covered
})
def test_lsymeig_AM(dtype, device, ashape, mshape, method):
    torch.manual_seed(seed)
    mata = torch.rand(ashape, dtype=dtype, device=device)
    matm = torch.rand(mshape, dtype=dtype, device=device) + \
        torch.eye(mshape[-1], dtype=dtype, device=device)  # make sure it's not singular
    mata = mata + mata.transpose(-2, -1).conj()
    matm = matm + matm.transpose(-2, -1).conj()
    mata = mata.requires_grad_()
    matm = matm.requires_grad_()
    linopa = LinearOperator.m(mata, is_hermitian=True)
    linopm = LinearOperator.m(matm, is_hermitian=True)
    fwd_options = {"method": method}

    na = ashape[-1]
    bshape = get_bcasted_dims(ashape[:-2], mshape[:-2])
    for neig in [2, ashape[-1]]:
        eigvals, eigvecs = lsymeig(linopa, M=linopm, neig=neig, **fwd_options)  # eigvals: (..., neig)
        eigvals = eigvals.to(eigvecs.dtype)
        assert list(eigvals.shape) == list([*bshape, neig])
        assert list(eigvecs.shape) == list([*bshape, na, neig])

        ax = linopa.mm(eigvecs)
        mxe = linopm.mm(torch.matmul(eigvecs, torch.diag_embed(eigvals, dim1=-2, dim2=-1)))
        assert torch.allclose(ax, mxe)

        # only perform gradcheck if neig is full, to reduce the computational cost
        if neig == ashape[-1]:
            def lsymeig_fcn(amat, mmat):
                # symmetrize
                amat = (amat + amat.transpose(-2, -1).conj()) * 0.5
                mmat = (mmat + mmat.transpose(-2, -1).conj()) * 0.5
                alinop = LinearOperator.m(amat, is_hermitian=True)
                mlinop = LinearOperator.m(mmat, is_hermitian=True)
                eigvals_, eigvecs_ = lsymeig(alinop, M=mlinop, neig=neig, **fwd_options)
                return eigvals_, eigvecs_.abs()

            gradcheck(lsymeig_fcn, (mata, matm))
            gradgradcheck(lsymeig_fcn, (mata, matm))

@device_dtype_float_test(only64=True, additional_kwargs={
    "shape": [(1000, 1000), (2, 1000, 1000), (2, 3, 1000, 1000)],
    "method": ["davidson"],  # list the methods here
    "mode": ["uppermost", "lowest"],
})
def test_symeig_A_large_methods(dtype, device, shape, method, mode):
    torch.manual_seed(seed)

    class ALarge(LinearOperator):
        def __init__(self, shape, dtype, device):
            super(ALarge, self).__init__(shape,
                                         is_hermitian=True,
                                         dtype=dtype,
                                         device=device)
            na = shape[-1]
            self.b = torch.arange(na, dtype=dtype, device=device).repeat(*shape[:-2], 1)

        def _mv(self, x):
            # x: (*BX, na)
            xb = x * self.b
            xsmall = x * 1e-3
            xp1 = torch.roll(xsmall, shifts=1, dims=-1)
            xm1 = torch.roll(xsmall, shifts=-1, dims=-1)
            return xb + xp1 + xm1

        def _getparamnames(self, prefix=""):
            return [prefix + "b"]

    neig = 2
    na = shape[-1]
    linop1 = ALarge(shape, dtype=dtype, device=device)
    fwd_options = {"method": method, "min_eps": 1e-8}

    # eigvals: (..., neig), eigvecs: (..., na, neig)
    eigvals, eigvecs = symeig(linop1, mode=mode, neig=neig, **fwd_options)

    # the matrix's eigenvalues will be around arange(na)
    if mode == "lowest":
        assert (eigvals < neig * 2).all()
    elif mode == "uppermost":
        assert (eigvals > na - neig * 2).all()

    assert list(eigvecs.shape) == list([*linop1.shape[:-1], neig])
    assert list(eigvals.shape) == list([*linop1.shape[:-2], neig])

    ax = linop1.mm(eigvecs)
    xe = torch.matmul(eigvecs, torch.diag_embed(eigvals, dim1=-2, dim2=-1))
    assert torch.allclose(ax, xe)

@device_dtype_float_test(only64=True, include_complex=True, additional_kwargs={
    "eivaloffset": [0, -4],
    "method": ["exacteig", "custom_exacteig"]
})
def test_symeig_A_degenerate(dtype, device, eivaloffset, method):
    # test if the gradient can be stably propagated if the loss function
    # does not depend on which degenerate eigenvectors are used
    # (note: the variable is changed in a certain way so that the degeneracy
    # is kept)

    torch.manual_seed(seed)
    n = 5
    neig = 3
    kwargs = {
        "dtype": dtype,
        "device": device,
    }
    print(dtype)
    # random matrix to be orthogonalized for the eigenvectors
    mat = torch.randn((n, n), **kwargs).requires_grad_()

    # matrix for the loss function
    P2 = torch.randn((n, n), **kwargs).requires_grad_()

    # the degenerate eigenvalues
    a = (torch.tensor([1.0, 2.0, 3.0], **kwargs) + eivaloffset)
    if torch.is_complex(a):
        a = a.real
    a = a.requires_grad_()

    bck_options = {
        "method": "exactsolve",
    }

    def get_loss(a, mat, P2):
        # get the orthogonal vector for the eigenvectors
        P, _ = torch.linalg.qr(mat)

        # line up the eigenvalues
        b = torch.cat((a[:2], a[1:2], a[2:], a[2:])).to(dtype)

        # construct the matrix
        diag = torch.diag_embed(b)
        A = torch.matmul(torch.matmul(P.transpose(-2, -1).conj(), diag), P)
        Alinop = LinearOperator.m(A, is_hermitian=True)

        eivals, eivecs = symeig(
            Alinop, neig=neig,
            method=method,
            bck_options=bck_options)
        U = eivecs[:, 1:3]  # the degenerate eigenvectors

        loss = torch.einsum("rc,rc->", torch.matmul(P2, U), U.conj())
        return loss

    gradcheck(get_loss, (a, mat, P2))
    gradgradcheck(get_loss, (a, mat, P2))

@device_dtype_float_test(only64=True, include_complex=True, additional_kwargs={
    "method": ["exacteig", "custom_exacteig"]
})
def test_symeig_AM_degenerate(dtype, device, method):
    # same as test_symeig_A_degenerate, but now with the overlap matrix

    torch.manual_seed(seed)
    n = 5
    neig = 3
    kwargs = {
        "dtype": dtype,
        "device": device,
    }
    # random matrix to be orthogonalized for the eigenvectors
    matA = torch.randn((n, n), **kwargs)
    matM = torch.rand((n, n), **kwargs)

    # matrix for the loss function
    P2 = torch.randn((n, n), **kwargs).requires_grad_()

    # the degenerate eigenvalues
    a = torch.tensor([1.0, 2.0, 3.0], **kwargs)
    if torch.is_complex(a):
        a = a.real
    a = a.requires_grad_()

    bck_options = {
        "method": "exactsolve",
    }

    def get_loss(a, matA, matM, P2):
        # get the orthogonal vector for the eigenvectors
        P, _ = torch.linalg.qr(matA)
        PM, _ = torch.linalg.qr(matM)

        # line up the eigenvalues
        b = torch.cat((a[:2], a[1:2], a[2:], a[2:])).to(dtype)

        # construct the matrix
        diag = torch.diag_embed(b)
        A = torch.matmul(torch.matmul(P.transpose(-2, -1).conj(), diag), P)
        M = torch.matmul(PM.transpose(-2, -1).conj(), PM)
        Alinop = LinearOperator.m(A, is_hermitian=True)
        Mlinop = LinearOperator.m(M, is_hermitian=True)

        eivals, eivecs = symeig(
            Alinop, M=Mlinop, neig=neig,
            method=method,
            bck_options=bck_options)
        U = eivecs[:, 1:3]  # the degenerate eigenvectors

        loss = torch.einsum("rc,rc->", torch.matmul(P2, U), U.conj())
        return loss

    gradcheck(get_loss, (a, matA, matM, P2))
    gradgradcheck(get_loss, (a, matA, matM, P2))


@device_dtype_float_test(only64=True)
def test_symeig_A_degenerate_req_not_sat(dtype, device):
    # test if the degenerate gradient returns nan if the requirments are not satisfied

    torch.manual_seed(seed)
    n = 5
    neig = 3
    kwargs = {
        "dtype": dtype,
        "device": device,
    }
    # random matrix to be orthogonalized for the eigenvectors
    mat = torch.randn((n, n), **kwargs).requires_grad_()

    # the degenerate eigenvalues
    a = torch.tensor([1.0, 2.0, 3.0], **kwargs).requires_grad_()
    bck_options = {
        "method": "exactsolve",
    }

    def get_loss(a, mat):
        # get the orthogonal vector for the eigenvectors
        P, _ = torch.linalg.qr(mat)

        # line up the eigenvalues
        b = torch.cat((a[:2], a[1:2], a[2:], a[2:]))

        # construct the matrix
        diag = torch.diag_embed(b)
        A = torch.matmul(torch.matmul(P.T, diag), P)
        Alinop = LinearOperator.m(A)

        eivals, eivecs = symeig(
            Alinop, neig=neig,
            method="custom_exacteig",
            bck_options=bck_options)
        U = eivecs[:, :3]  # the degenerate eigenvectors are in 1,2
        loss = torch.sum(U ** 4)
        return loss

    with warnings.catch_warnings(record=True) as w, enable_debug():
        loss = get_loss(a, mat)
        loss.backward()
        assert len(w) == 1
        wmsg = str(w[0].message).lower()
        assert "degener" in wmsg
        assert "loss function" in wmsg
        assert "incorrect" in wmsg
        assert w[0].category == MathWarning

############## svd #############
@device_dtype_float_test(only64=True, include_complex=True, additional_kwargs={
    "shape": [(4, 3), (2, 1, 3, 4)],
    "method": ["exacteig", "custom_exacteig"],
})
def test_svd_A(dtype, device, shape, method):
    torch.manual_seed(seed)
    mat1 = torch.rand(shape, dtype=dtype, device=device)
    mat1 = mat1.requires_grad_()
    linop1 = LinearOperator.m(mat1, is_hermitian=False)
    fwd_options = {"method": method}

    min_mn = min(shape[-1], shape[-2])
    for k in [min_mn]:
        u, s, vh = svd(linop1, k=k, **fwd_options)  # u: (..., m, k), s: (..., k), vh: (..., k, n)
        assert list(u.shape) == list([*linop1.shape[:-1], k])
        assert list(s.shape) == list([*linop1.shape[:-2], k])
        assert list(vh.shape) == list([*linop1.shape[:-2], k, linop1.shape[-1]])

        keye = torch.zeros((*shape[:-2], k, k), dtype=dtype, device=device) + \
            torch.eye(k, dtype=dtype, device=device)
        assert torch.allclose(u.transpose(-2, -1).conj() @ u, keye)
        assert torch.allclose(vh @ vh.transpose(-2, -1).conj(), keye)
        if k == min_mn:
            assert torch.allclose(mat1, u @ torch.diag_embed(s.to(u.dtype)) @ vh)

        def svd_fcn(amat, only_s=False):
            alinop = LinearOperator.m(amat, is_hermitian=False)
            u_, s_, vh_ = svd(alinop, k=k, **fwd_options)
            if only_s:
                return s_
            else:
                return u_.abs(), s_, vh_.abs()

        gradcheck(svd_fcn, (mat1,))
        gradgradcheck(svd_fcn, (mat1,))

############## solve ##############
@device_dtype_float_test()
def test_solve_nonsquare_err(dtype, device):
    torch.manual_seed(seed)
    mat = torch.rand((3, 2), dtype=dtype, device=device)
    mat2 = torch.rand((3, 3), dtype=dtype, device=device)
    linop = LinearOperator.m(mat)
    linop2 = LinearOperator.m(mat2)
    B = torch.rand((3, 1), dtype=dtype, device=device)

    try:
        res = solve(linop, B)
        assert False, "A RuntimeError must be raised if the A linear operator in solve not square"
    except RuntimeError:
        pass

    try:
        res = solve(linop2, B, M=linop)
        assert False, "A RuntimeError must be raised if the M linear operator in solve is not square"
    except RuntimeError:
        pass

@device_dtype_float_test()
def test_solve_mismatch_err(dtype, device):
    torch.manual_seed(seed)
    shapes = [
        #   A      B      M
        ([(3, 3), (2, 1), (3, 3)], "the B shape does not match with A"),
        ([(3, 3), (3, 2), (2, 2)], "the M shape does not match with A"),
    ]
    for (ashape, bshape, mshape), msg in shapes:
        amat = torch.rand(ashape, dtype=dtype, device=device)
        bmat = torch.rand(bshape, dtype=dtype, device=device)
        mmat = torch.rand(mshape, dtype=dtype, device=device) + \
            torch.eye(mshape[-1], dtype=dtype, device=device)
        amat = amat + amat.transpose(-2, -1)
        mmat = mmat + mmat.transpose(-2, -1)

        alinop = LinearOperator.m(amat)
        mlinop = LinearOperator.m(mmat)
        try:
            res = solve(alinop, B=bmat, M=mlinop)
            assert False, "A RuntimeError must be raised if %s" % msg
        except RuntimeError:
            pass

@device_dtype_float_test(only64=True, include_complex=True, additional_kwargs={
    "ashape": [(2, 2), (2, 2, 2), (2, 1, 2, 2)],
    "bshape": [(2, 2), (2, 2, 2), (2, 1, 2, 2)],
    "method": ["exactsolve", "custom_exactsolve"],
    "hermit": [False, True],
})
def test_solve_A(dtype, device, ashape, bshape, method, hermit):
    torch.manual_seed(seed)
    na = ashape[-1]
    checkgrad = method.endswith("exactsolve") and len(ashape) == len(bshape) == 2

    ncols = bshape[-1] - 1
    bshape = [*bshape[:-1], ncols]
    xshape = list(get_bcasted_dims(ashape[:-2], bshape[:-2])) + [na, ncols]
    fwd_options = {"method": method, "min_eps": 1e-9}
    bck_options = {"method": method}

    amat = torch.rand(ashape, dtype=dtype, device=device) * 0.1 + \
        torch.eye(ashape[-1], dtype=dtype, device=device)
    bmat = torch.rand(bshape, dtype=dtype, device=device)

    amat = amat.requires_grad_()
    bmat = bmat.requires_grad_()

    def prepare(amat):
        if hermit:
            return (amat + amat.transpose(-2, -1).conj()) * 0.5
        return amat

    def solvefcn(amat, bmat):
        # is_hermitian=hermit is required to force the hermitian status in numerical gradient
        alinop = LinearOperator.m(prepare(amat), is_hermitian=hermit)
        x = solve(A=alinop, B=bmat,
                  **fwd_options,
                  bck_options=bck_options)
        return x

    x = solvefcn(amat, bmat)
    assert list(x.shape) == xshape

    ax = LinearOperator.m(prepare(amat)).mm(x)
    assert torch.allclose(ax, bmat)

    if checkgrad:
        gradcheck(solvefcn, (amat, bmat))
        gradgradcheck(solvefcn, (amat, bmat))

@device_dtype_float_test(only64=True, include_complex=True, additional_kwargs={
    "method": ["scipy_gmres", "broyden1", "cg", "bicgstab", "gmres"],
})
def test_solve_A_methods(dtype, device, method):

    if dtype in [torch.complex128, torch.complex64]:
        if method in ["scipy_gmres", "gmres"]:
            pytest.xfail("%s does not work for complex input" % method)

    torch.manual_seed(seed)
    na = 100
    ashape = (na, na)
    bshape = (2, na, na)
    options = {
        "scipy_gmres": {},
        "broyden1": {
            "alpha": -0.2,
        },
        "cg": {
            "rtol": 1e-8  # stringent rtol required to meet the torch.allclose tols
        },
        "bicgstab": {
            "rtol": 1e-8,
        },
        "gmres": {}
    }[method]
    fwd_options = {"method": method, **options}

    ncols = bshape[-1] - 1
    bshape = [*bshape[:-1], ncols]
    xshape = list(get_bcasted_dims(ashape[:-2], bshape[:-2])) + [na, ncols]

    amat = torch.rand(ashape, dtype=dtype, device=device) * 0.1 + \
        torch.eye(ashape[-1], dtype=dtype, device=device)
    bmat = torch.rand(bshape, dtype=dtype, device=device) + 0.1
    amat = (amat + amat.transpose(-2, -1).conj()) * 0.5

    amat = amat.requires_grad_()
    bmat = bmat.requires_grad_()

    def solvefcn(amat, bmat):
        alinop = LinearOperator.m(amat)
        x = solve(A=alinop, B=bmat, **fwd_options)
        return x

    x = solvefcn(amat, bmat)
    assert list(x.shape) == xshape

    ax = LinearOperator.m(amat).mm(x)

    assert torch.allclose(ax, bmat)

@device_dtype_float_test(only64=True, include_complex=True, additional_kwargs={
    "ashape": [(2, 2), (2, 2, 2), (2, 1, 2, 2)],
    "bshape": [(2, 2), (2, 2, 2), (2, 1, 2, 2)],
    "eshape": [(2, 2), (2, 2, 2), (2, 1, 2, 2)],
    "method": ["exactsolve", "custom_exactsolve"],
})
def test_solve_AE(dtype, device, ashape, bshape, eshape, method):
    torch.manual_seed(seed)
    na = ashape[-1]

    # save time by enabling gradchecker only on some cases
    checkgrad = method.endswith("exactsolve") and len(ashape) == len(bshape) == len(eshape) == 2

    ncols = bshape[-1] - 1
    bshape = [*bshape[:-1], ncols]
    eshape = [*eshape[:-2], ncols]
    xshape = list(get_bcasted_dims(ashape[:-2], bshape[:-2], eshape[:-1])) + [na, ncols]
    fwd_options = {"method": method}
    bck_options = {"method": method}

    amat = torch.rand(ashape, dtype=dtype, device=device) * 0.1 + \
        torch.eye(ashape[-1], dtype=dtype, device=device)
    bmat = torch.rand(bshape, dtype=dtype, device=device)
    emat = torch.rand(eshape, dtype=dtype, device=device)

    amat = amat.requires_grad_()
    bmat = bmat.requires_grad_()
    emat = emat.requires_grad_()

    def solvefcn(amat, bmat, emat):
        alinop = LinearOperator.m(amat)
        x = solve(A=alinop, B=bmat, E=emat,
                  **fwd_options,
                  bck_options=bck_options)
        return x

    x = solvefcn(amat, bmat, emat)
    assert list(x.shape) == xshape

    ax = LinearOperator.m(amat).mm(x)
    xe = torch.matmul(x, torch.diag_embed(emat, dim2=-1, dim1=-2))

    assert torch.allclose(ax - xe, bmat)

    if checkgrad:
        gradcheck(solvefcn, (amat, bmat, emat))
        gradgradcheck(solvefcn, (amat, bmat, emat))

@device_dtype_float_test(only64=True, include_complex=True, additional_kwargs={
    "abeshape": [(2, 2), (2, 2, 2), (2, 1, 2, 2)],
    "mshape": [(2, 2), (2, 2, 2), (2, 1, 2, 2)],
    "method": ["exactsolve", "custom_exactsolve"],
})
def test_solve_AEM(dtype, device, abeshape, mshape, method):
    torch.manual_seed(seed)
    na = abeshape[-1]
    ashape = abeshape
    bshape = abeshape
    eshape = abeshape

    # save time by enabling gradchecker only on some cases
    checkgrad = method.endswith("exactsolve") and len(abeshape) == len(mshape) == 2

    ncols = bshape[-1] - 1
    bshape = [*bshape[:-1], ncols]
    eshape = [*eshape[:-2], ncols]
    xshape = list(get_bcasted_dims(ashape[:-2], bshape[:-2], eshape[:-1], mshape[:-2])) + [na, ncols]
    fwd_options = {"method": method, "min_eps": 1e-9}
    bck_options = {"method": method}  # exactsolve at backward just to test the forward solve

    amat = torch.rand(ashape, dtype=dtype, device=device) * 0.1 + \
        torch.eye(ashape[-1], dtype=dtype, device=device)
    mmat = torch.rand(mshape, dtype=dtype, device=device) * 0.1 + \
        torch.eye(mshape[-1], dtype=dtype, device=device) * 0.5
    bmat = torch.rand(bshape, dtype=dtype, device=device)
    emat = torch.rand(eshape, dtype=dtype, device=device)
    mmat = (mmat + mmat.transpose(-2, -1).conj()) * 0.5

    amat = amat.requires_grad_()
    mmat = mmat.requires_grad_()
    bmat = bmat.requires_grad_()
    emat = emat.requires_grad_()

    def solvefcn(amat, mmat, bmat, emat):
        mmat = (mmat + mmat.transpose(-2, -1).conj()) * 0.5
        alinop = LinearOperator.m(amat)
        mlinop = LinearOperator.m(mmat, is_hermitian=True)
        x = solve(A=alinop, B=bmat, E=emat, M=mlinop,
                  **fwd_options,
                  bck_options=bck_options)
        return x

    x = solvefcn(amat, mmat, bmat, emat)
    assert list(x.shape) == xshape

    ax = LinearOperator.m(amat).mm(x)
    mxe = LinearOperator.m(mmat).mm(torch.matmul(x, torch.diag_embed(emat.to(x.dtype), dim2=-1, dim1=-2)))
    y = ax - mxe
    assert torch.allclose(y, bmat)

    # gradient checker
    if checkgrad:
        gradcheck(solvefcn, (amat, mmat, bmat, emat))
        gradgradcheck(solvefcn, (amat, mmat, bmat, emat))

@device_dtype_float_test(only64=True, include_complex=True, additional_kwargs={
    "method": ["broyden1", "cg", "bicgstab"],
})
def test_solve_AEM_methods(dtype, device, method):
    torch.manual_seed(seed)
    na = 100
    nc = na // 2

    amshape = (na, na)
    eshape = (nc,)
    bshape = (2, na, nc)
    options = {
        "scipy_gmres": {},
        "broyden1": {
            "alpha": -0.2,
        },
        "cg": {
            "rtol": 1e-8  # stringent rtol required to meet the torch.allclose tols
        },
        "bicgstab": {},
    }[method]
    fwd_options = {"method": method, **options}

    amat = torch.rand(amshape, dtype=dtype, device=device) * 0.1 + \
        torch.eye(amshape[-1], dtype=dtype, device=device)
    mmat = torch.rand(amshape, dtype=dtype, device=device) * 0.05 + \
        torch.eye(amshape[-1], dtype=dtype, device=device) * 0.5
    bmat = torch.rand(bshape, dtype=dtype, device=device) + 0.1
    emat = torch.rand(eshape, dtype=dtype, device=device) * 0.1
    amat = (amat + amat.transpose(-2, -1).conj()) * 0.5
    mmat = (mmat + mmat.transpose(-2, -1).conj()) * 0.5

    amat = amat.requires_grad_()
    bmat = bmat.requires_grad_()
    emat = emat.requires_grad_()

    def solvefcn(amat, bmat, emat, mmat):
        alinop = LinearOperator.m(amat)
        mlinop = LinearOperator.m(mmat)
        x = solve(A=alinop, B=bmat, E=emat, M=mlinop, **fwd_options)
        return x

    x = solvefcn(amat, bmat, emat, mmat)
    ax = LinearOperator.m(amat).mm(x)
    mxe = LinearOperator.m(mmat).mm(x) @ torch.diag_embed(emat)
    assert torch.allclose(ax - mxe, bmat)



import torch
from xitorch._core.pure_function import get_pure_function

class NNModule(torch.nn.Module):
    def __init__(self, a):
        super(NNModule, self).__init__()
        self.a = a  # must be a param

    def forward(self, x):
        return self.a * x

class NNModule0(torch.nn.Module):
    def __init__(self):
        super(NNModule0, self).__init__()

    def forward(self, x):
        return 2 * x


dtype = torch.float64

def test_wrap_nnmodule():
    a = torch.nn.Parameter(torch.tensor([1.2], dtype=dtype))
    x = torch.tensor([1.4], dtype=dtype, requires_grad=True)

    modules = [NNModule(a), NNModule0()]
    nparams = [2, 1]

    for i in range(len(modules)):
        module = modules[i]
        pfcn = get_pure_function(module)
        objparams = pfcn.objparams()
        assert len(objparams) == nparams[i] - 1  # 1 for the parameter x
        if len(objparams) == 1:
            assert objparams[0] is a



import xitorch

def test_debug_default():
    assert not xitorch.is_debug_enabled()

def test_set_debug_mode():
    xitorch.set_debug_mode(True)
    assert xitorch.is_debug_enabled()
    xitorch.set_debug_mode(False)
    assert not xitorch.is_debug_enabled()

def test_debug_context_manager():
    assert not xitorch.is_debug_enabled()
    with xitorch.enable_debug():
        assert xitorch.is_debug_enabled()
        with xitorch.disable_debug():
            assert not xitorch.is_debug_enabled()
        assert xitorch.is_debug_enabled()
    assert not xitorch.is_debug_enabled()



import torch
import inspect
from typing import Callable, List, Tuple, Union, Sequence
from xitorch._utils.attr import set_attr, del_attr
from xitorch._utils.unique import Uniquifier
from xitorch._core.editable_module import EditableModule
from contextlib import contextmanager
from abc import abstractmethod

__all__ = ["get_pure_function", "make_sibling"]

############################ functional ###############################
class PureFunction(object):
    """
    PureFunction class wraps methods to make it stateless and expose the pure
    function to take inputs of the original inputs (`params`) and the object's
    states (`objparams`).
    For functions, this class only acts as a thin wrapper.
    """

    def __init__(self, fcntocall: Callable):
        self._state_change_allowed = True
        self._allobjparams = self._get_all_obj_params_init()
        self._uniq = Uniquifier(self._allobjparams)
        self._cur_objparams = self._uniq.get_unique_objs()
        self._fcntocall = fcntocall

        # restore stack stores list of (objparams, identical)
        # everytime the objparams are set, it will store the old objparams
        # and indication if the old and new objparams are identical
        self._restore_stack: List[Tuple[List, bool]] = []

    def __call__(self, *params):
        return self._fcntocall(*params)

    @abstractmethod
    def _get_all_obj_params_init(self):
        pass

    @abstractmethod
    def _set_all_obj_params(self, allobjparams):
        pass

    def objparams(self) -> List:
        return self._cur_objparams

    def set_objparams(self, objparams: List):
        # TODO: check if identical with current object parameters
        identical = _check_identical_objs(objparams, self._cur_objparams)
        self._restore_stack.append((self._cur_objparams, identical))
        if not identical:
            allobjparams = self._uniq.map_unique_objs(objparams)
            self._set_all_obj_params(allobjparams)
            self._cur_objparams = list(objparams)

    def restore_objparams(self):
        old_objparams, identical = self._restore_stack.pop(-1)
        if not identical:
            allobjparams = self._uniq.map_unique_objs(old_objparams)
            self._set_all_obj_params(allobjparams)
            self._cur_objparams = old_objparams

    @contextmanager
    def useobjparams(self, objparams: List):
        if not self._state_change_allowed:
            raise RuntimeError("The state change is disabled")
        try:
            self.set_objparams(objparams)
            yield
        finally:
            self.restore_objparams()

    @contextmanager
    def disable_state_change(self):
        try:
            prev_status = self._state_change_allowed
            self._state_change_allowed = False
            yield
        finally:
            self._state_change_allowed = prev_status

class FunctionPureFunction(PureFunction):
    def _get_all_obj_params_init(self):
        return []

    def _set_all_obj_params(self, objparams):
        pass

class EditableModulePureFunction(PureFunction):
    def __init__(self, obj: EditableModule, method: Callable):
        self.obj = obj
        self.method = method
        super().__init__(method)

    def _get_all_obj_params_init(self) -> List:
        return list(self.obj.getparams(self.method.__name__))

    def _set_all_obj_params(self, allobjparams: List):
        self.obj.setparams(self.method.__name__, *allobjparams)

class TorchNNPureFunction(PureFunction):
    def __init__(self, obj: torch.nn.Module, method: Callable):
        self.obj = obj
        self.method = method
        super().__init__(method)

    def _get_all_obj_params_init(self) -> List:
        # get the tensors in the torch.nn.Module to be used as params
        named_params = list(self.obj.named_parameters())
        if len(named_params) == 0:
            paramnames: List[str] = []
            obj_params: List[Union[torch.Tensor, torch.nn.Parameter]] = []
        else:
            paramnames_temp, obj_params_temp = zip(*named_params)
            paramnames = list(paramnames_temp)
            obj_params = list(obj_params_temp)
        self.names = paramnames
        return obj_params

    def _set_all_obj_params(self, objparams: List):
        for (name, param) in zip(self.names, objparams):
            del_attr(self.obj, name)  # delete required in case the param is not a torch.nn.Parameter
            set_attr(self.obj, name, param)

class SingleSiblingPureFunction(PureFunction):
    def __init__(self, fcn: Callable, fcntocall: Callable):
        self.pfunc = get_pure_function(fcn)
        super().__init__(fcntocall)

    def _get_all_obj_params_init(self) -> List:
        return self.pfunc._get_all_obj_params_init()

    def _set_all_obj_params(self, allobjparams: List):
        self.pfunc._set_all_obj_params(allobjparams)

class MultiSiblingPureFunction(PureFunction):
    def __init__(self, fcns: Sequence[Callable], fcntocall: Callable):
        self.pfuncs = [get_pure_function(fcn) for fcn in fcns]
        self.npfuncs = len(self.pfuncs)
        super().__init__(fcntocall)

    def _get_all_obj_params_init(self) -> List:
        res: List[Union[torch.Tensor, torch.nn.Parameter]] = []
        self.cumsum_idx = [0] * (self.npfuncs + 1)
        for i, pfunc in enumerate(self.pfuncs):
            objparams = pfunc._get_all_obj_params_init()
            res = res + objparams
            self.cumsum_idx[i + 1] = self.cumsum_idx[i] + len(objparams)
        return res

    def _set_all_obj_params(self, allobjparams: List):
        for i, pfunc in enumerate(self.pfuncs):
            pfunc._set_all_obj_params(allobjparams[self.cumsum_idx[i]:self.cumsum_idx[i + 1]])

def _check_identical_objs(objs1: List, objs2: List) -> bool:
    for obj1, obj2 in zip(objs1, objs2):
        if id(obj1) != id(obj2):
            return False
    return True

def get_pure_function(fcn) -> PureFunction:
    """
    Get the pure function form of the function or method ``fcn``.

    Arguments
    ---------
    fcn: function or method
        Function or method to be converted into a ``PureFunction`` by exposing
        the hidden parameters affecting its outputs.

    Returns
    -------
    PureFunction
        The pure function wrapper
    """

    errmsg = "The input function must be a function, a method of " \
        "torch.nn.Module, a method of xitorch.EditableModule, or a sibling method"

    if isinstance(fcn, PureFunction):
        return fcn

    elif inspect.isfunction(fcn) or isinstance(fcn, torch.jit.ScriptFunction):
        return FunctionPureFunction(fcn)

    # if it is a method from an object, unroll the parameters and add
    # the object's parameters as well
    elif inspect.ismethod(fcn) or hasattr(fcn, "__call__"):
        if inspect.ismethod(fcn):
            obj = fcn.__self__
        else:
            obj = fcn
            fcn = fcn.__call__

        if isinstance(obj, EditableModule):
            return EditableModulePureFunction(obj, fcn)
        elif isinstance(obj, torch.nn.Module):
            return TorchNNPureFunction(obj, fcn)
        else:
            raise RuntimeError(errmsg)

    else:
        raise RuntimeError(errmsg)

def make_sibling(*pfuncs) -> Callable[[Callable], PureFunction]:
    """
    Used as a decor to mark the decorated function as a sibling method of the
    input ``pfunc``.
    Sibling method is a method that is virtually belong to the same object, but
    behaves differently.
    Changing the state of the decorated function will also change the state of
    ``pfunc`` and its other siblings.
    """
    if len(pfuncs) == 0:
        raise TypeError("At least 1 function is required as the argument")
    elif len(pfuncs) == 1:
        return lambda fcn: SingleSiblingPureFunction(pfuncs[0], fcntocall=fcn)
    else:
        return lambda fcn: MultiSiblingPureFunction(pfuncs, fcntocall=fcn)



from contextlib import contextmanager

__all__ = ["is_debug_enabled", "set_debug_mode", "enable_debug", "disable_debug"]

class DebugSingleton(object):
    class __DebugSingleton:
        def __init__(self):
            self._isdebug = False  # default mode is not in the debug mode

        def set_debug_mode(self, mode):
            self._isdebug = mode

        def get_debug_mode(self):
            return self._isdebug

    instance = None

    def __init__(self):
        if DebugSingleton.instance is None:
            DebugSingleton.instance = DebugSingleton.__DebugSingleton()

    def __getattr__(self, name):
        return getattr(self.instance, name)

    def __setattr__(self, name, val):
        return setattr(self.instance, name, val)

def set_debug_mode(mode):
    dbg_obj = DebugSingleton()
    dbg_obj.set_debug_mode(mode)

def is_debug_enabled():
    dbg_obj = DebugSingleton()
    return dbg_obj.get_debug_mode()

@contextmanager
def enable_debug():
    try:
        dbg_mode = is_debug_enabled()
        set_debug_mode(True)
        yield
    except Exception as e:
        raise e
    finally:
        set_debug_mode(dbg_mode)

@contextmanager
def disable_debug():
    try:
        dbg_mode = is_debug_enabled()
        set_debug_mode(False)
        yield
    except Exception as e:
        raise e
    finally:
        set_debug_mode(dbg_mode)



from typing import List
import time
import torch
from xitorch.integrate import solve_ivp

class RemoveTimeArg(torch.nn.Module):
    def __init__(self, module: torch.nn.Module):
        super().__init__()
        self._module = module

    def forward(self, t: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return self._module(y)

def test_ivp_speed():
    # test the speed of ivp compared to manually implement it
    dydt_module = torch.nn.Sequential(
        torch.nn.Linear(3, 10),
        torch.nn.LogSigmoid(),
        torch.nn.Linear(10, 10),
        torch.nn.LogSigmoid(),
        torch.nn.Linear(10, 3),
    )

    def manual_euler(dydt_module: torch.nn.Module, ts: torch.Tensor, y0: torch.Tensor):
        # ts: (nt,)
        # y0: (nb, ndim)
        y_lst: List[torch.Tensor] = []
        y_lst.append(y0)
        y = y0
        for i in range(1, ts.size(0)):
            dydt = dydt_module(y)  # (nb, ndim)
            y = y + dydt * (ts[i] - ts[i - 1])
            y_lst.append(y)
        yres = torch.stack(y_lst, dim=0)
        return yres

    dydt_module2 = RemoveTimeArg(dydt_module)
    ts = torch.linspace(0, 1.0, 100).requires_grad_()
    y0 = torch.randn((5, 3)).requires_grad_()

    t0 = time.time()
    for i in range(10):
        yt0 = manual_euler(dydt_module, ts, y0)
    t1 = time.time()
    for i in range(10):
        yt1 = solve_ivp(dydt_module2, ts, y0, method="euler")
    t2 = time.time()
    for i in range(10):
        _ = torch.autograd.grad(yt0, (ts, y0), torch.ones_like(yt0), retain_graph=True)
    t3 = time.time()
    for i in range(10):
        _ = torch.autograd.grad(yt1, (ts, y0), torch.ones_like(yt1), retain_graph=True)
    t4 = time.time()

    # forward timing
    dt1 = t1 - t0  # manual
    dt2 = t2 - t1  # xitorch
    assert dt2 < 1.5 * dt1  # cannot be much slower than the manual implementation

    # backward timing
    dtg1 = t3 - t2  # manual
    dtg2 = t4 - t3  # xitorch
    assert dtg2 < 15.0 * dtg1  # it is expected to be slower than manual implementation

if __name__ == "__main__":
    test_ivp_speed()



import torch
from torch.autograd import gradcheck, gradgradcheck
from xitorch.grad.jachess import jac, hess
from xitorch._core.editable_module import EditableModule
from xitorch._core.pure_function import get_pure_function

dtype = torch.float64

def func1(A, b, x0):
    x = torch.matmul(A, x0) + b
    x = torch.nn.Softplus()(x)
    return x

def hfunc1(A, b, x0):
    x = torch.matmul(A, x0) + b
    x = torch.nn.Softplus()(x)
    return x.sum()

def scalar_func(A, b, x0):
    x = A * x0 + b
    x = torch.nn.Softplus()(x)
    return x

class func2(EditableModule):
    def __init__(self, b):
        self.b = b

    def getparamnames(self, methodname, prefix=""):
        if methodname == "__call__":
            return [prefix + "b"]
        else:
            raise KeyError("Params for method %s cannot be found" % methodname)

    def __call__(self, A, b, x0):
        x = torch.matmul(A * self.b, x0) + b
        x = torch.nn.Softplus()(x)
        return x

class hfunc2(EditableModule):
    def __init__(self, b):
        self.b = b

    def getparamnames(self, methodname, prefix=""):
        if methodname == "__call__":
            return [prefix + "b"]
        else:
            raise KeyError("Params for method %s cannot be found" % methodname)

    def __call__(self, A, b, x0):
        x = torch.matmul(A * self.b, x0) + b
        x = torch.nn.Softplus()(x)
        return x.sum()

def getfnscalarparams():
    A = torch.rand((), dtype=dtype, requires_grad=True)
    b = torch.rand((), dtype=dtype, requires_grad=True)
    x = torch.rand((), dtype=dtype, requires_grad=True)
    return (A, b, x)

def getfnparams(na):
    A = torch.rand((na, na), dtype=dtype, requires_grad=True)
    b = torch.rand((na, 1), dtype=dtype, requires_grad=True)
    x = torch.rand((na, 1), dtype=dtype, requires_grad=True)
    return (A, b, x)

def getnnparams(na):
    b = torch.rand((na, 1), dtype=dtype, requires_grad=True)
    return (b,)

def test_jac_func():
    na = 3
    params = getfnparams(na)
    nnparams = getnnparams(na)
    nparams = len(params)
    all_idxs = [None, (0,), (1,), (0, 1), (0, 1, 2)]
    funcs = [func1, func2(*nnparams)]

    for func in funcs:
        for idxs in all_idxs:
            if idxs is None:
                gradparams = params
            else:
                gradparams = [params[i] for i in idxs]

            jacs = jac(func, params, idxs=idxs)
            assert len(jacs) == len(gradparams)

            y = func(*params)
            nout = torch.numel(y)
            nins = [torch.numel(p) for p in gradparams]
            v = torch.rand_like(y).requires_grad_()
            for i in range(len(jacs)):
                assert list(jacs[i].shape) == [nout, nins[i]]

            # get rmv
            jacs_rmv = torch.autograd.grad(y, gradparams, grad_outputs=v, create_graph=True)
            # the jac LinearOperator has shape of (nout, nin), so we need to flatten v
            jacs_rmv0 = [jc.rmv(v.view(-1)) for jc in jacs]

            # calculate the mv
            w = [torch.rand_like(p) for p in gradparams]
            jacs_lmv = [torch.autograd.grad(jacs_rmv[i], (v,), grad_outputs=w[i], retain_graph=True)[
                0] for i in range(len(jacs))]
            jacs_lmv0 = [jacs[i].mv(w[i].view(-1)) for i in range(len(jacs))]

            for i in range(len(jacs)):
                assert torch.allclose(jacs_rmv[i].view(-1), jacs_rmv0[i].view(-1))
                assert torch.allclose(jacs_lmv[i].view(-1), jacs_lmv0[i].view(-1))

def test_jac_grad():
    na = 3
    params = getfnparams(na)
    params2 = [torch.rand(1, dtype=dtype).requires_grad_() for p in params]
    jacs = jac(func1, params)
    nout = jacs[0].shape[-2]

    def fcnr(i, v, *params):
        jacs = jac(func1, params)
        return jacs[i].rmv(v.view(-1))

    def fcnl(i, w, *params):
        jacs = jac(func1, params)
        return jacs[i].mv(w.view(-1))

    def fcnr2(i, v, *params2):
        fmv = get_pure_function(jacs[i].rmv)
        params0 = v.view(-1)
        params1 = fmv.objparams()
        params12 = [p1 * p2 for p1, p2 in zip(params1, params2)]
        with fmv.useobjparams(params12):
            return fmv(params0)

    def fcnl2(i, w, *params2):
        fmv = get_pure_function(jacs[i].mv)
        params0 = w.view(-1)
        params1 = fmv.objparams()
        params12 = [p1 * p2 for (p1, p2) in zip(params1, params2)]
        with fmv.useobjparams(params12):
            return fmv(params0)

    v = torch.rand((na,), dtype=dtype, requires_grad=True)
    w = [torch.rand_like(p).requires_grad_() for p in params]
    for i in range(len(jacs)):
        gradcheck(fcnr, (i, v, *params))
        gradgradcheck(fcnr, (i, v, *params))
        gradcheck(fcnl, (i, w[i], *params))
        gradgradcheck(fcnl, (i, w[i], *params))
        gradcheck(fcnr2, (i, v, *params2))
        gradgradcheck(fcnr2, (i, v, *params2))
        gradcheck(fcnl2, (i, w[i], *params2))
        gradgradcheck(fcnl2, (i, w[i], *params2))

def test_jac_method_grad():
    na = 3
    params = getfnparams(na)
    nnparams = getnnparams(na)
    num_nnparams = len(nnparams)
    jacs = jac(func2(*nnparams), params)
    nout = jacs[0].shape[-2]

    def fcnr(i, v, *allparams):
        nnparams = allparams[:num_nnparams]
        params = allparams[num_nnparams:]
        jacs = jac(func2(*nnparams), params)
        return jacs[i].rmv(v.view(-1))

    def fcnl(i, v, *allparams):
        nnparams = allparams[:num_nnparams]
        params = allparams[num_nnparams:]
        jacs = jac(func2(*nnparams), params)
        return jacs[i].mv(v.view(-1))

    v = torch.rand((na,), dtype=dtype, requires_grad=True)
    w = [torch.rand_like(p).requires_grad_() for p in params]
    for i in range(len(jacs)):
        gradcheck(fcnr, (i, v, *nnparams, *params))
        gradgradcheck(fcnr, (i, v, *nnparams, *params))
        gradcheck(fcnl, (i, w[i], *nnparams, *params))
        gradgradcheck(fcnl, (i, w[i], *nnparams, *params))

def test_jac_scalar_func():
    params = getfnscalarparams()
    nparams = len(params)
    all_idxs = [None, (0,), (1,), (0, 1), (0, 1, 2)]
    func = scalar_func

    for idxs in all_idxs:
        if idxs is None:
            gradparams = params
        else:
            gradparams = [params[i] for i in idxs]

        jacs = jac(func, params, idxs=idxs)
        assert len(jacs) == len(gradparams)

        y = func(*params)
        nout = torch.numel(y)
        nins = [torch.numel(p) for p in gradparams]
        v = torch.rand_like(y).requires_grad_()
        for i in range(len(jacs)):
            assert list(jacs[i].shape) == [nout, nins[i]]

        # get rmv
        jacs_rmv = torch.autograd.grad(y, gradparams, grad_outputs=v, create_graph=True)
        # the jac LinearOperator has shape of (nout, nin), so we need to flatten v
        jacs_rmv0 = [jc.rmv(v.view(-1)) for jc in jacs]

        # calculate the mv
        w = [torch.rand_like(p) for p in gradparams]
        jacs_lmv = [torch.autograd.grad(jacs_rmv[i], (v,), grad_outputs=w[i], retain_graph=True)[0]
                    for i in range(len(jacs))]
        jacs_lmv0 = [jacs[i].mv(w[i].view(-1)) for i in range(len(jacs))]

        for i in range(len(jacs)):
            assert torch.allclose(jacs_rmv[i].view(-1), jacs_rmv0[i].view(-1))
            assert torch.allclose(jacs_lmv[i].view(-1), jacs_lmv0[i].view(-1))

def test_hess_func():
    na = 3
    params = getfnparams(na)
    nnparams = getnnparams(na)
    nparams = len(params)
    all_idxs = [None, (0,), (1,), (0, 1), (0, 1, 2)]
    funcs = [hfunc1, hfunc2(*nnparams)]

    for func in funcs:
        for idxs in all_idxs:
            if idxs is None:
                gradparams = params
            else:
                gradparams = [params[i] for i in idxs]

            hs = hess(func, params, idxs=idxs)
            assert len(hs) == len(gradparams)

            y = func(*params)
            nins = [torch.numel(p) for p in gradparams]
            w = [torch.rand_like(p) for p in gradparams]
            for i in range(len(hs)):
                assert list(hs[i].shape) == [nins[i], nins[i]]

            # assert the values
            dfdy = torch.autograd.grad(y, gradparams, create_graph=True)
            hs_mv_man = [torch.autograd.grad(dfdy[i], (gradparams[i],), grad_outputs=w[i],
                                             retain_graph=True)[0] for i in range(len(dfdy))]
            hs_mv = [hs[i].mv(w[i].reshape(-1, nins[i])) for i in range(len(dfdy))]
            for i in range(len(dfdy)):
                assert torch.allclose(hs_mv[i].view(-1), hs_mv_man[i].view(-1))

def test_hess_grad():
    na = 3
    params = getfnparams(na)
    params2 = [torch.rand(1, dtype=dtype).requires_grad_() for p in params]
    hs = hess(hfunc1, params)

    def fcnl(i, v, *params):
        hs = hess(hfunc1, params)
        return hs[i].mv(v.view(-1))

    def fcnl2(i, v, *params2):
        fmv = get_pure_function(hs[i].mv)
        params0 = v.view(-1)
        params1 = fmv.objparams()
        params12 = [p1 * p2 for (p1, p2) in zip(params1, params2)]
        with fmv.useobjparams(params12):
            return fmv(params0)

    w = [torch.rand_like(p).requires_grad_() for p in params]
    for i in range(len(hs)):
        gradcheck(fcnl, (i, w[i], *params))
        gradgradcheck(fcnl, (i, w[i], *params))
        gradcheck(fcnl2, (i, w[i], *params2))
        gradgradcheck(fcnl2, (i, w[i], *params2))



import inspect
import warnings
import functools

def deprecated(date_str):
    return lambda obj: _deprecated(obj, date_str)

def _deprecated(obj, date_str):
    if inspect.isfunction(obj):
        name = "Function %s" % (obj.__str__())
    elif inspect.isclass(obj):
        name = "Class %s" % (obj.__name__)

    if inspect.ismethod(obj) or inspect.isfunction(obj):
        @functools.wraps(obj)
        def fcn(*args, **kwargs):
            warnings.warn(
                "%s is deprecated since %s" % (name, date_str),
                stacklevel=2)
            return obj(*args, **kwargs)
        return fcn

    elif inspect.isclass(obj):
        # replace the __init__ function
        old_init = obj.__init__

        @functools.wraps(old_init)
        def newinit(*args, **kwargs):
            warnings.warn(
                "%s is deprecated since %s" % (name, date_str),
                stacklevel=2)
            return old_init(*args, **kwargs)
        obj.__init__ = newinit
        return obj



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['xitorch',
 'xitorch._core',
 'xitorch._docstr',
 'xitorch._impls',
 'xitorch._impls.integrate',
 'xitorch._impls.integrate.ivp',
 'xitorch._impls.integrate.mcsamples',
 'xitorch._impls.interpolate',
 'xitorch._impls.linalg',
 'xitorch._impls.optimize',
 'xitorch._impls.optimize.root',
 'xitorch._tests',
 'xitorch._utils',
 'xitorch.debug',
 'xitorch.grad',
 'xitorch.integrate',
 'xitorch.interpolate',
 'xitorch.linalg',
 'xitorch.optimize']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.24.1', 'scipy>=1.10.0', 'torch>=1.13.1']

setup_kwargs = {
    'name': 'xitorch',
    'version': '0.0.0',
    'description': 'Differentiable scientific computing library',
    'long_description': '# `xitorch`: differentiable scientific computing library\n\n![Build](https://img.shields.io/github/workflow/status/xitorch/xitorch/CI)\n[![Docs](https://img.shields.io/readthedocs/xitorch)](https://xitorch.readthedocs.io/)\n[![Code coverage](https://img.shields.io/codecov/c/github/xitorch/xitorch)](https://codecov.io/gh/xitorch/xitorch)\n\n`xitorch` is a PyTorch-based library of differentiable functions and functionals that\ncan be widely used in scientific computing applications as well as deep learning.\n\nThe documentation can be found at: https://xitorch.readthedocs.io/\n\n## Example\n\nFinding root of a function:\n\n```python\nimport torch\nfrom xitorch.optimize import rootfinder\n\ndef func1(y, A):  # example function\n    return torch.tanh(A @ y + 0.1) + y / 2.0\n\n# set up the parameters and the initial guess\nA = torch.tensor([[1.1, 0.4], [0.3, 0.8]]).requires_grad_()\ny0 = torch.zeros((2, 1))  # zeros as the initial guess\n\n# finding a root\nyroot = rootfinder(func1, y0, params=(A,))\n\n# calculate the derivatives\ndydA, = torch.autograd.grad(yroot.sum(), (A,), create_graph=True)\ngrad2A, = torch.autograd.grad(dydA.sum(), (A,), create_graph=True)\n```\n\n## Modules\n\n* [`linalg`](xitorch/linalg/): Linear algebra and sparse linear algebra module\n* [`optimize`](xitorch/optimize/): Optimization and root finder module\n* [`integrate`](xitorch/integrate/): Quadrature and integration module\n* [`interpolate`](xitorch/interpolate/): Interpolation\n\n## Requirements\n\n* python >=3.8.1,<3.12\n* pytorch 1.13.1 or higher (install [here](https://pytorch.org/))\n\n## Getting started\n\nAfter fulfilling all the requirements, type the commands below to install `xitorch`\n\n    python -m pip install xitorch\n\nOr to install from GitHub:\n\n    python -m pip install git+https://github.com/xitorch/xitorch.git\n\nFinally, if you want to make an editable install from source:\n\n    git clone https://github.com/xitorch/xitorch.git\n    cd xitorch\n    python -m pip install -e .\n\nNote that the last option is only available per [PEP 660](https://peps.python.org/pep-0660/), so you will require [pip >= 23.1](https://pip.pypa.io/en/stable/news/#v21-3)\n    \n## Used in\n\n* Differentiable Quantum Chemistry (DQC): https://dqc.readthedocs.io/\n\n## Gallery\n\nNeural mirror design ([example 01](examples/01-mirror-design/)):\n\n![neural mirror design](examples/01-mirror-design/images/mirror.gif)\n\nInitial velocity optimization in molecular dynamics ([example 02](examples/02-molecular-dynamics/)):\n\n![molecular dynamics](examples/02-molecular-dynamics/images/md.gif)\n',
    'author': 'mfkasim1',
    'author_email': 'firman.kasim@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://xitorch.readthedocs.io/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.1,<3.12',
}


setup(**setup_kwargs)



import random
import torch
import numpy as np
from functorch import vmap
from torch.autograd import gradcheck, gradgradcheck
import xitorch as xt
from xitorch.integrate import quad, solve_ivp, mcquad, SQuad
from xitorch._tests.utils import device_dtype_float_test

################################## quadrature ##################################
class IntegrationNNModule(torch.nn.Module):
    # cos(a*x + b * c)
    def __init__(self, a, b):
        super(IntegrationNNModule, self).__init__()
        self.a = a
        self.b = b

    def forward(self, x, c):
        return torch.cos(self.a * x + self.b * c)

class IntegrationModule(xt.EditableModule):
    # cos(a*x + b * c)
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def forward(self, x, c):
        return torch.cos(self.a * x + self.b * c)

    def getparamnames(self, methodname, prefix=""):
        return [prefix + "a", prefix + "b"]

class IntegrationNNMultiModule(torch.nn.Module):
    # cos(a*x + b * c), sin(a*x + b*c)
    def __init__(self, a, b):
        super(IntegrationNNMultiModule, self).__init__()
        self.a = a
        self.b = b

    def forward(self, x, c):
        return torch.cos(self.a * x + self.b * c), torch.sin(self.a * x + self.b * c)

class IntegrationMultiModule(xt.EditableModule):
    # cos(a*x + b * c), sin(a*x + b*c)
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def forward(self, x, c):
        return torch.cos(self.a * x + self.b * c), torch.sin(self.a * x + self.b * c)

    def getparamnames(self, methodname, prefix=""):
        return [prefix + "a", prefix + "b"]

class IntegrationInfModule(torch.nn.Module):
    def __init__(self, w):
        super(IntegrationInfModule, self).__init__()
        self.w = w

    def forward(self, x):
        return torch.exp(-x * x / (2 * self.w * self.w))

@device_dtype_float_test(only64=True, additional_kwargs={
    "clss": [IntegrationModule, IntegrationNNModule],
})
def test_quad(dtype, device, clss):
    torch.manual_seed(100)
    random.seed(100)
    nr = 2
    fwd_options = {
        "method": "leggauss",
        "n": 100,
    }

    a = torch.nn.Parameter(torch.rand((nr,), dtype=dtype, device=device).requires_grad_())
    b = torch.nn.Parameter(torch.randn((nr,), dtype=dtype, device=device).requires_grad_())
    c = torch.randn((nr,), dtype=dtype, device=device).requires_grad_()
    xl = torch.zeros((1,), dtype=dtype, device=device).requires_grad_()
    xu = (torch.ones((1,), dtype=dtype, device=device) * 0.5).requires_grad_()

    module = clss(a, b)
    y = quad(module.forward, xl, xu, params=(c,), **fwd_options)
    ytrue = (torch.sin(a * xu + b * c) - torch.sin(a * xl + b * c)) / a
    assert torch.allclose(y, ytrue)

    def getloss(a, b, c, xl, xu):
        module = clss(a, b)
        y = quad(module.forward, xl, xu, params=(c,), **fwd_options)
        return y

    gradcheck(getloss, (a, b, c, xl, xu))
    gradgradcheck(getloss, (a, b, c, xl, xu))
    # check if not all parameters require grad
    gradcheck(getloss, (a, b.detach(), c, xl, xu))

@device_dtype_float_test(only64=True, additional_kwargs={
    "clss": [IntegrationMultiModule, IntegrationNNMultiModule],
})
def test_quad_multi(dtype, device, clss):
    torch.manual_seed(100)
    random.seed(100)
    nr = 4
    fwd_options = {
        "method": "leggauss",
        "n": 100,
    }

    a = torch.nn.Parameter(torch.rand((nr,), dtype=dtype, device=device).requires_grad_())
    b = torch.nn.Parameter(torch.randn((nr,), dtype=dtype, device=device).requires_grad_())
    c = torch.randn((nr,), dtype=dtype, device=device).requires_grad_()
    xl = torch.zeros((1,), dtype=dtype, device=device).requires_grad_()
    xu = (torch.ones((1,), dtype=dtype, device=device) * 0.5).requires_grad_()

    module = clss(a, b)
    y = quad(module.forward, xl, xu, params=(c,), **fwd_options)
    ytrue0 = (torch.sin(a * xu + b * c) - torch.sin(a * xl + b * c)) / a
    ytrue1 = (-torch.cos(a * xu + b * c) + torch.cos(a * xl + b * c)) / a
    assert len(y) == 2
    assert torch.allclose(y[0], ytrue0)
    assert torch.allclose(y[1], ytrue1)

@device_dtype_float_test(only64=True, additional_kwargs={
    "totensor": [True, False]
})
def test_quad_inf(dtype, device, totensor):
    torch.manual_seed(100)
    random.seed(100)
    nr = 4
    fwd_options = {
        "method": "leggauss",
        "n": 100,
    }
    w = torch.nn.Parameter(torch.abs(torch.randn((nr,), dtype=dtype, device=device)).requires_grad_())

    if totensor:
        xl = torch.tensor(-float("inf"), dtype=dtype, device=device)
        xu = torch.tensor(float("inf"), dtype=dtype, device=device)
    else:
        xl = -float("inf")
        xu = float("inf")

    def get_loss(w):
        module = IntegrationInfModule(w)
        y = quad(module.forward, xl, xu, params=[], **fwd_options)
        return y

    y = get_loss(w)
    ytrue = w * np.sqrt(2 * np.pi)
    assert torch.allclose(y, ytrue)
    if totensor:
        gradcheck(get_loss, (w,))
        gradgradcheck(get_loss, (w,))

################################## ivp ##################################
class IVPNNModule(torch.nn.Module):
    # dydt: -a * y * t - b * y - c * y
    def __init__(self, a, b):
        super(IVPNNModule, self).__init__()
        self.a = a
        self.b = b

    def forward(self, t, y, c):
        return -self.a * y * t - self.b * y - c * y

class IVPModule(xt.EditableModule):
    # dydt: -a * y * t - b * y - c * y
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def forward(self, t, y, c):
        return -self.a * y * t - self.b * y - c * y

    def getparamnames(self, methodname, prefix=""):
        return [prefix + "a", prefix + "b"]

@device_dtype_float_test(only64=True, additional_kwargs={
    "clss": [IVPModule, IVPNNModule],
})
def test_ivp(dtype, device, clss):
    torch.manual_seed(100)
    random.seed(100)
    nr = 2
    nt = 5
    t0 = 0.0
    t1 = 0.2
    fwd_options = {
        "method": "rk4",
    }

    a = torch.nn.Parameter(torch.rand((nr,), dtype=dtype, device=device).requires_grad_())
    b = torch.nn.Parameter(torch.randn((nr,), dtype=dtype, device=device).requires_grad_())
    c = torch.randn((nr,), dtype=dtype, device=device).requires_grad_()
    ts = torch.linspace(t0, t1, nt, dtype=dtype, device=device).requires_grad_()
    y0 = torch.rand((nr,), dtype=dtype, device=device).requires_grad_()
    ts1 = ts.unsqueeze(-1)

    def getoutput(a, b, c, ts, y0):
        module = clss(a, b)
        yt = solve_ivp(module.forward, ts, y0, params=(c,), **fwd_options)
        return yt

    yt = getoutput(a, b, c, ts, y0)
    yt_true = y0 * torch.exp(-(0.5 * a * (ts1 + t0) + b + c) * (ts1 - t0))
    assert torch.allclose(yt, yt_true)

    gradcheck(getoutput, (a, b, c, ts, y0))
    gradgradcheck(getoutput, (a, b, c, ts, y0))

@device_dtype_float_test(only64=True, additional_kwargs={
    "method_tol": [
        ("rk4", (1e-8, 1e-5)),
        ("rk38", (1e-8, 1e-5)),
        ("rk45", (1e-8, 1e-5)),
        ("rk23", (1e-6, 1e-4)),
        ("euler", (5e-2, 1e-4)),  # yes, don't use euler method
    ],
    "clss": [IVPModule, IVPNNModule],
})
def test_ivp_methods(dtype, device, method_tol, clss):
    torch.manual_seed(100)
    random.seed(100)
    nr = 2
    nb = 3  # batch dimension
    nt = 5
    t0 = 0.0
    t1 = 0.2
    a = torch.nn.Parameter(torch.rand((nr,), dtype=dtype, device=device).requires_grad_())
    b = torch.nn.Parameter(torch.randn((nr,), dtype=dtype, device=device).requires_grad_())
    c = torch.randn((nr,), dtype=dtype, device=device).requires_grad_()
    ts = torch.linspace(t0, t1, nt, dtype=dtype, device=device).requires_grad_()
    y0 = torch.rand((nb, nr), dtype=dtype, device=device).requires_grad_()
    ts1 = ts.unsqueeze(-1).unsqueeze(-1)

    method, (rtol, atol) = method_tol
    fwd_options = {
        "method": method,
    }

    def getoutput(a, b, c, ts, y0):
        module = clss(a, b)
        yt = solve_ivp(module.forward, ts, y0, params=(c,), **fwd_options)
        return yt

    yt = getoutput(a, b, c, ts, y0)
    yt_true = y0 * torch.exp(-(0.5 * a * (ts1 + t0) + b + c) * (ts1 - t0))
    assert torch.allclose(yt, yt_true, rtol=rtol, atol=atol)

# Renamed it to "atest" to avoid it being tested.
# This is due to functorch's capability for operating with torch.autograd.Function was retracted in torch 1.13
# because it's silently produces wrong results: https://github.com/pytorch/functorch/issues/207
# This capability will resume in torch 2.0, but we still have to rewrite our _SolveIVP method
# See https://pytorch.org/docs/master/notes/extending.func.html
@device_dtype_float_test(only64=True, additional_kwargs={
    "method_tol": [
        ("rk4", (1e-8, 1e-5)),
        ("rk38", (1e-8, 1e-5)),
        # ("rk45", (1e-8, 1e-5)),
        # ("rk23", (1e-6, 1e-4)),
        ("euler", (5e-2, 1e-4)),  # yes, don't use euler method
    ],
    "clss": [IVPModule, IVPNNModule],
})
def atest_ivp_methods_batch(dtype, device, method_tol, clss):
    # test batched using vmap
    torch.manual_seed(100)
    random.seed(100)
    nr = 2
    nb = 3  # batch, only for y
    nbatch = 4  # batch, for both t and y
    nt = 5
    t0 = 0.0
    t1 = 0.2
    a = torch.nn.Parameter(torch.rand((nr,), dtype=dtype, device=device).requires_grad_())
    b = torch.nn.Parameter(torch.randn((nr,), dtype=dtype, device=device).requires_grad_())
    c = torch.randn((nr,), dtype=dtype, device=device).requires_grad_()
    ts = torch.linspace(t0, t1, nt, dtype=dtype, device=device)[..., None].expand(-1, nbatch)
    ts = ts.requires_grad_()  # (nt, nbatch)
    y0 = torch.rand((nbatch, nb, nr), dtype=dtype, device=device).requires_grad_()
    ts1 = ts[..., None, None]  # (nt, nbatch, 1, 1)

    method, (rtol, atol) = method_tol
    fwd_options = {
        "method": method,
    }
    solve_ivp_batch = vmap(solve_ivp, in_dims=(None, 1, 0, (None,)), out_dims=1)

    def getoutput(a, b, c, ts, y0):
        module = clss(a, b)
        yt = solve_ivp_batch(module.forward, ts, y0, (c,), **fwd_options)
        return yt

    yt = getoutput(a, b, c, ts, y0)
    assert yt.shape == (nt, nbatch, nb, nr)
    yt_true = y0 * torch.exp(-(0.5 * a * (ts1 + t0) + b + c) * (ts1 - t0))
    assert torch.allclose(yt, yt_true, rtol=rtol, atol=atol)

################################## mcquad ##################################
class MCQuadLogProbNNModule(torch.nn.Module):
    def __init__(self, w):
        super(MCQuadLogProbNNModule, self).__init__()
        self.w = w

    def forward(self, x):
        # x, w are single-element tensors
        return -x * x / (2 * self.w * self.w)

class MCQuadFcnModule(xt.EditableModule):
    def __init__(self, a):
        self.a = a

    def forward(self, x):
        # return self.a*self.a * x * x
        return torch.exp(-x * x / (2 * self.a * self.a))

    def getparamnames(self, methodname, prefix=""):
        return [prefix + "a"]

def get_true_output(w, a):
    # return a*a*w*w
    return 1.0 / torch.sqrt(1 + w * w / (a * a))

@device_dtype_float_test(only64=True, additional_kwargs={
    "method": ["mh", "_dummy1d"],
})
def test_mcquad(dtype, device, method):
    torch.manual_seed(100)
    random.seed(100)

    w = torch.nn.Parameter(torch.tensor(1.2, dtype=dtype, device=device))
    a = torch.tensor(0.3, dtype=dtype, device=device).requires_grad_()
    x0 = torch.tensor(0.0, dtype=dtype, device=device)

    if method == "mh":
        fwd_options = {
            "method": "mh",
            "step_size": 0.6,
            "nsamples": 10000,
            "nburnout": 2,
        }
    else:
        # using deterministic forward method just to check the backward operation
        fwd_options = {
            "method": "_dummy1d",
            "nsamples": 100,
            "lb": -float("inf"),
            "ub": float("inf"),
        }

    def getoutput(w, a, x0):
        logp = MCQuadLogProbNNModule(w)
        fcn = MCQuadFcnModule(a)
        res = mcquad(fcn.forward, logp.forward, x0, fparams=[], pparams=[], **fwd_options)
        return res

    rtol = 2e-2 if method != "_dummy1d" else 1e-3
    epf = getoutput(w, a, x0)
    epf_true = get_true_output(w, a)
    assert torch.allclose(epf, epf_true, rtol=rtol)

    # skip gradient check if it is not the deterministic method
    if method != "_dummy1d":
        return

    # manually check the gradient
    g = torch.tensor(0.7, dtype=dtype, device=device).reshape(epf.shape).requires_grad_()
    ga, gw      = torch.autograd.grad(epf, (a, w), grad_outputs=g, create_graph=True)
    # different implementation
    ga2, gw2     = torch.autograd.grad(epf, (a, w), grad_outputs=g, retain_graph=True, create_graph=False)
    ga_true, gw_true = torch.autograd.grad(epf_true, (a, w), grad_outputs=g, create_graph=True)
    assert torch.allclose(gw, gw_true)
    assert torch.allclose(ga, ga_true)
    assert torch.allclose(gw2, gw_true)
    assert torch.allclose(ga2, ga_true)

    ggaw, ggaa, ggag      = torch.autograd.grad(ga, (w, a, g), retain_graph=True, allow_unused=True)
    ggaw_true, ggaa_true, ggag_true = torch.autograd.grad(ga_true, (w, a, g), retain_graph=True, allow_unused=True)
    print("ggaw", ggaw, ggaw_true, (ggaw - ggaw_true) / ggaw_true)
    print("ggaa", ggaa, ggaa_true, (ggaa - ggaa_true) / ggaa_true)
    print("ggag", ggag, ggag_true, (ggag - ggag_true) / ggag_true)
    assert torch.allclose(ggaa, ggaa_true)
    assert torch.allclose(ggag, ggag_true)
    ggww, ggwa, ggwg      = torch.autograd.grad(gw, (w, a, g), allow_unused=True)
    ggww_true, ggwa_true, ggwg_true = torch.autograd.grad(gw_true, (w, a, g), allow_unused=True)
    print("ggwa", ggwa, ggwa_true, (ggwa - ggwa_true) / ggwa_true)
    print("ggwg", ggwg, ggwg_true, (ggwg - ggwg_true) / ggwg_true)
    print("ggww", ggww, ggww_true, (ggww - ggww_true) / ggww_true)
    assert torch.allclose(ggwa, ggwa_true)
    assert torch.allclose(ggwg, ggwg_true)
    assert torch.allclose(ggww, ggww_true)

################################## SQuad ##################################
@device_dtype_float_test(only64=True, additional_kwargs={
    "imethod": list(enumerate(["trapz", "cspline"])),
})
def test_squad(dtype, device, imethod):
    x = torch.tensor([0.0, 1.0, 2.0, 4.0, 5.0, 7.0],
                     dtype=dtype, device=device).requires_grad_()
    y = torch.tensor([[1.0, 2.0, 2.0, 1.5, 1.2, 4.0],
                      [0.0, 0.8, 1.0, 1.5, 2.0, 1.4]],
                     dtype=dtype, device=device).requires_grad_()

    # true values
    ycumsum_trapz = torch.tensor(  # obtained by calculating manually
        [[0.0, 1.5, 3.5, 7.0, 8.35, 13.55],
         [0.0, 0.4, 1.3, 3.8, 5.55, 8.95]],
        dtype=dtype, device=device)
    ycspline_natural = torch.tensor(  # obtained using scipy's CubicSpline and quad
        [[0.0, 1.5639104372355428, 3.6221791255289135, 7.2068053596614945, 8.4994887166897, 13.11119534565217],
         [0.0, 0.43834626234132584, 1.3733074753173484, 3.724083215796897, 5.494693230049832, 9.181717209378409]],
        dtype=dtype, device=device)

    i, method = imethod
    option = [{}, {"bc_type": "natural"}][i]
    ytrue = [ycumsum_trapz, ycspline_natural][i]

    def getval(x, y, tpe):
        quad = SQuad(x, method=method, **option)
        if tpe == "cumsum":
            return quad.cumsum(y, dim=-1)
        else:
            return quad.integrate(y, dim=-1)

    # getparamnames
    quad = SQuad(x, method=method, **option)
    quad.assertparams(quad.cumsum, y, dim=-1)
    quad.assertparams(quad.integrate, y, dim=-1)

    # cumsum
    ycumsum = getval(x, y, "cumsum")
    assert torch.allclose(ycumsum, ytrue)

    # integrate
    yintegrate = getval(x, y, "integrate")
    assert torch.allclose(yintegrate, ytrue[..., -1])

    gradcheck(getval, (x, y, "cumsum"))
    gradgradcheck(getval, (x, y, "cumsum"), nondet_tol=1e-6)
    gradcheck(getval, (x, y, "integrate"))
    gradgradcheck(getval, (x, y, "integrate"), nondet_tol=1e-6)


if __name__ == "__main__":
    # with torch.autograd.detect_anomaly():
    test_mcquad()



import random
import torch
from xitorch import LinearOperator, EditableModule
from xitorch.linalg import lsymeig, solve
from xitorch.optimize import rootfinder, equilibrium, minimize
from xitorch._utils.bcast import get_bcasted_dims
from xitorch._tests.utils import device_dtype_float_test, assert_no_memleak

# memory leak tests

seed = 12345

############## linear algebra functions ##############
@device_dtype_float_test(only64=True, onlycpu=True, additional_kwargs={
    "ashape": [(3, 3)],
    "mshape": [(3, 3)],
    "method": ["custom_exacteig"],
})
def test_lsymeig_AM_mem(dtype, device, ashape, mshape, method):

    def _test_lsymeig():
        torch.manual_seed(seed)
        mata = torch.rand(ashape, dtype=dtype, device=device)
        matm = torch.rand(mshape, dtype=dtype, device=device) + \
            torch.eye(mshape[-1], dtype=dtype, device=device)  # make sure it's not singular
        mata = mata + mata.transpose(-2, -1)
        matm = matm + matm.transpose(-2, -1)
        mata = mata.requires_grad_()
        matm = matm.requires_grad_()
        fwd_options = {"method": method}

        na = ashape[-1]
        bshape = get_bcasted_dims(ashape[:-2], mshape[:-2])
        neig = ashape[-1]

        def lsymeig_fcn(amat, mmat):
            # symmetrize
            amat = (amat + amat.transpose(-2, -1)) * 0.5
            mmat = (mmat + mmat.transpose(-2, -1)) * 0.5
            alinop = LinearOperator.m(amat, is_hermitian=True)
            mlinop = LinearOperator.m(mmat, is_hermitian=True)
            eigvals_, eigvecs_ = lsymeig(alinop, M=mlinop, neig=neig, **fwd_options)
            return eigvals_, eigvecs_

        eival, eivec = lsymeig_fcn(mata, matm)
        loss = (eival * eival).sum() + (eivec * eivec).sum()
        # using autograd.grad instead of .backward because backward has a known
        # memory leak problem
        grads = torch.autograd.grad(loss, (mata, matm), create_graph=True)

    assert_no_memleak(_test_lsymeig)

@device_dtype_float_test(only64=True, additional_kwargs={
    "abeshape": [(2, 2)],
    "mshape": [(2, 2)],
    "method": ["custom_exactsolve"],
})
def test_solve_AEM_mem(dtype, device, abeshape, mshape, method):
    def _test_solve():
        torch.manual_seed(seed)
        na = abeshape[-1]
        ashape = abeshape
        bshape = abeshape
        eshape = abeshape
        checkgrad = method.endswith("exactsolve")

        ncols = bshape[-1] - 1
        bshape = [*bshape[:-1], ncols]
        eshape = [*eshape[:-2], ncols]
        xshape = list(get_bcasted_dims(ashape[:-2], bshape[:-2], eshape[:-1], mshape[:-2])) + [na, ncols]
        fwd_options = {"method": method, "min_eps": 1e-9}
        bck_options = {"method": method}  # exactsolve at backward just to test the forward solve

        amat = torch.rand(ashape, dtype=dtype, device=device) * 0.1 + \
            torch.eye(ashape[-1], dtype=dtype, device=device)
        mmat = torch.rand(mshape, dtype=dtype, device=device) * 0.1 + \
            torch.eye(mshape[-1], dtype=dtype, device=device) * 0.5
        bmat = torch.rand(bshape, dtype=dtype, device=device)
        emat = torch.rand(eshape, dtype=dtype, device=device)
        mmat = (mmat + mmat.transpose(-2, -1)) * 0.5

        amat = amat.requires_grad_()
        mmat = mmat.requires_grad_()
        bmat = bmat.requires_grad_()
        emat = emat.requires_grad_()

        def solvefcn(amat, mmat, bmat, emat):
            mmat = (mmat + mmat.transpose(-2, -1)) * 0.5
            alinop = LinearOperator.m(amat)
            mlinop = LinearOperator.m(mmat)
            x = solve(A=alinop, B=bmat, E=emat, M=mlinop,
                      **fwd_options,
                      bck_options=bck_options)
            return x

        loss = (solvefcn(amat, mmat, bmat, emat) ** 2).sum()
        # using autograd.grad instead of .backward because backward has a known
        # memory leak problem
        grads = torch.autograd.grad(loss, (amat, mmat, bmat, emat), create_graph=True)

    assert_no_memleak(_test_solve)

############## optimize functions ##############
class DummyModule(EditableModule):
    def __init__(self, a, sumoutput=False):
        super().__init__()
        self.a = a
        self.sumoutput = sumoutput

    def forward(self, x):
        y = self.a ** 2 * x ** 4 - 1
        if self.sumoutput:
            return y.sum()
        else:
            return y

    def getparamnames(self, methodname, prefix=""):
        return [prefix + "a"]

@device_dtype_float_test(only64=True, onlycpu=True)
def test_rootfinder_mem(dtype, device):
    def _test_rf():
        clss = DummyModule  # only need to test it with one clsas
        torch.manual_seed(100)
        random.seed(100)

        nbatch = 2000
        fwd_options = {
            "method": "broyden1",
            "f_tol": 1e-9,
            "alpha": -0.5,
        }

        a = (torch.ones((nbatch,), dtype=dtype) + 0.5).requires_grad_()
        y0 = torch.ones((nbatch,), dtype=dtype)

        def getloss(a):
            model = clss(a)
            y = rootfinder(model.forward, y0, **fwd_options)
            return y

        loss = (getloss(a) ** 2).sum()
        # using torch.autograd.grad instead of backward because backward
        # has a memory leak problem
        grads = torch.autograd.grad(loss, (a,), create_graph=True)

    assert_no_memleak(_test_rf)

@device_dtype_float_test(only64=True, onlycpu=True)
def test_equil_mem(dtype, device):
    def _test_equil():
        clss = DummyModule
        torch.manual_seed(100)
        random.seed(100)

        nbatch = 2000
        fwd_options = {
            "method": "broyden1",
            "f_tol": 1e-9,
            "alpha": -0.5,
        }

        a = (torch.ones((nbatch,), dtype=dtype) + 0.5).requires_grad_()
        y0 = torch.ones((nbatch,), dtype=dtype)

        def getloss(a):
            model = clss(a)
            y = equilibrium(model.forward, y0, **fwd_options)
            return y

        loss = (getloss(a) ** 2).sum()
        grads = torch.autograd.grad(loss, (a,), create_graph=True)

    assert_no_memleak(_test_equil)

@device_dtype_float_test(only64=True, onlycpu=True)
def test_minimize_mem(dtype, device):
    def _test_minimize():
        clss = DummyModule
        torch.manual_seed(400)
        random.seed(100)

        nbatch = 2000
        fwd_options = {
            "method": "broyden1",
            "f_tol": 1e-9,
            "alpha": -0.5,
        }

        a = (torch.ones((nbatch,), dtype=dtype) + 0.5).requires_grad_()
        y0 = torch.ones((nbatch,), dtype=dtype)

        def getloss(a):
            model = clss(a, sumoutput=True)
            y = minimize(model.forward, y0, **fwd_options)
            return y

        loss = (getloss(a) ** 2).sum()
        grads = torch.autograd.grad(loss, (a,), create_graph=True)

    assert_no_memleak(_test_minimize)
